import logging
import unittest

from omegaml import Omega
from omegaml.store.logging import OmegaSimpleLogger, OmegaLoggingHandler
from omegaml.tests.util import OmegaTestMixin


class OmegaLoggingTests(OmegaTestMixin, unittest.TestCase):
    def setUp(self):
        self.om = Omega()
        self.clean()
        self._reset_log()

    def tearDown(self):
        self.assertTrue(all(m.name is not None for m in self.om.datasets.list(raw=True)))

    def _reset_log(self):
        self.om.datasets.drop('.omega/logs', force=True)

    def test_simple_logger(self):
        logger = self.om.logger
        self.assertIsInstance(logger, OmegaSimpleLogger)
        logger.setLevel('DEBUG')
        logger.info('info message')
        logger.error('error message')
        logger.debug('debug message')
        logger.warning('warning message')
        logger.critical('critical message')
        for level in OmegaSimpleLogger.levels:
            expected = 1 if level != 'QUIET' else 0
            df = logger.dataset.get(level=level)
            self.assertTrue(len(df) == expected, 'level {} did not have a record'.format(level))
            if expected:
                # msg contains the exact message as passed in
                self.assertEqual(df.iloc[0].msg, '{} message'.format(level).lower())
                # text contains the formatted msg
                self.assertNotEqual(df.iloc[0].text, '{} message'.format(level).lower())
                self.assertIn('{} message'.format(level).lower(), df.iloc[0].text)

    def test_simple_logger_level(self):
        logger = self.om.logger
        self.assertIsInstance(logger, OmegaSimpleLogger)
        logger.setLevel('INFO')
        logger.info('info message')
        logger.debug('error message')
        df = logger.dataset.get(level='DEBUG')
        self.assertTrue(len(df) == 0)
        df = logger.dataset.get(level='INFO')
        self.assertTrue(len(df) == 1)

    def test_simple_logger_quiet(self):
        logger = self.om.logger
        self.assertIsInstance(logger, OmegaSimpleLogger)
        logger.info('initialize') # need this to initialize
        logger.setLevel('QUIET')
        logger.error('info message')
        logger.debug('error message')
        df = logger.dataset.get(level='INFO')
        self.assertTrue(len(df) == 1)
        df = logger.dataset.get(level='DEBUG')
        self.assertTrue(len(df) == 0)
        df = logger.dataset.get(level='ERROR')
        self.assertTrue(len(df) == 0)

    def test_loghandler(self):
        pylogger = logging.getLogger()
        omlogger = self.om.logger
        handler = OmegaLoggingHandler.setup(logger=pylogger, level='DEBUG')
        pylogger.setLevel('DEBUG')
        pylogger.info('info message')
        pylogger.error('error message')
        pylogger.warning('warning message')
        pylogger.debug('debug message')
        df = omlogger.dataset.get()
        self.assertEqual(len(df), 5) # including init
        pylogger.handlers.remove(handler)
        # msg contains the exact message as passed in
        for level in ['INFO', 'ERROR', 'WARNING', 'DEBUG']:
            df = omlogger.dataset.get(filter=dict(level=level, logger='root'))
            expected = 1
            self.assertEqual(len(df), expected, 'expected 1 message for level {}'.format(level))
            if expected:
                self.assertEqual(df.iloc[0].msg, '{} message'.format(level).lower())
                # text contains the formatted msg
                self.assertNotEqual(df.iloc[0].text, '{} message'.format(level).lower())
                self.assertIn('{} message'.format(level).lower(), df.iloc[0].text)


    def test_named_simplelogger(self):
        """
        test we can get a named logger
        """
        # get a named logger
        logger = self.om.logger.getLogger('foo')
        self.assertIsInstance(logger, OmegaSimpleLogger)
        self.assertNotEqual(logger, self.om.logger)
        self.assertNotEqual(logger.name, self.om.logger.name)
        logger.info('foo')
        df = logger.dataset.get(level='INFO')
        self.assertTrue(len(df) == 1)
        self.assertEqual(df.iloc[0]['logger'], 'foo')
        # change logger name of the default logger
        self.om.logger.name = 'myname'
        self.om.logger.info('foo')
        df = self.om.logger.dataset.get(filter=dict(level='INFO', logger='myname'))
        self.assertTrue(len(df) == 1)
        self.assertEqual(df.iloc[-1]['logger'], 'myname')

    def test_python_logger_capture(self):
        om = self.om
        # -- get a python logger and add message
        # -- expect message is not in om logger dataset
        logger = logging.getLogger('mytestlogger')
        logger.info('foo')
        df = om.logger.dataset.get(level='INFO')
        self.assertTrue(len(df) == 0)
        # -- capture python logger
        # -- expect message is in om logger dataset
        with om.logger.capture(logger):
            logger.info('foo')
        df = om.logger.dataset.get(level='INFO')
        self.assertTrue(len(df) == 1)




import string

import warnings

from apispec import APISpec
from apispec.ext.marshmallow import MarshmallowPlugin
from marshmallow import fields, Schema

from omegaml.util import markup


class SignatureMixin:
    """ Associate a models corresponding dataset

    This provides methods to record additional metadata on models.
    On runtime-called calls, this additional metadata is used to
    provide defaults for Xname, Yname, rName respectively.

    Usage::

        # recording information
        om.models.link_dataset('modelname', Xname=..., Yname=...)
        om.models.link_docs('modelname', doc_or_ref)

        # using default dataset names, using the metadata.attributes['dataset'] entry
        om.runtime.model('mymodel').fit('*', '*')
        om.runtime.model('mymodel').predict('*')

    Notes:
        - Provides pre/post processing for model actions. Currently fit(), predict()
        - Pre/post processing is triggered when called as
          ModelBackend.perform('method', *args, **kwargs)
    """

    # terms used in this class
    # -- "datatype" => marshmallow Schema class/instance
    # -- "schema" => "schemas" component

    def link_docs(self, name, doc_or_ref, **kwargs):
        """ set or link documentation returned by om.models.help()

        Args:
            name: the model name
            doc_or_ref: the docstring or the URL to a web page
            **kwargs:

        Returns:

        """
        meta = self.metadata(name, **kwargs)
        meta.attributes['docs'] = doc_or_ref
        return meta.save()

    def link_datatype(self, name, X=None, Y=None, result=None, orient=None, meta=None, actions=None, **kwargs):
        """

        Args:
            name (str): the name of the model, script
            X (Schema): the schema for the X data (dataX), applies to models, scripts
            Y (Schema): the schema for the Y data (dataY), applies to models only
            result (Schema): the schema for the result, applies to scripts only
            orient (str): the datatype orientation, defaults to 'columns' for models, 'records' for datasets and scripts
            actions (list): the list of valid actions ['predict', ...],
               or a list of tuples to map actions to http methods [('predict', ['post', ...]), ...]. If
               the http method is not specified, it is assumed as 'post'

        Returns:

        """
        meta = meta or self.metadata(name, **kwargs)
        if actions:
            if isinstance(actions, (list, tuple)) and isinstance(actions[0], str):
                # list of strings
                actions = {action: ['post'] for action in actions}
            else:
                # list of tuples [(http_method, [action, ...])] or dict
                actions = dict(actions)
        else:
            actions = None
        existing = meta.attributes.setdefault('signature', {})
        meta.attributes['signature'] = {
            'X': {
                'datatype': f'{X.__module__}.{X.__name__}',
                'schema': self._schema_from_datatype(X),
            } if X is not None else existing.get('X'),
            'Y': {
                'datatype': f'{Y.__module__}.{Y.__name__}',
                'schema': self._schema_from_datatype(Y),
            } if Y is not None else existing.get('Y'),
            'result': {
                'datatype': f'{result.__module__}.{result.__name__}',
                'schema': self._schema_from_datatype(result),
            } if result is not None else existing.get('result'),
            'actions': actions,
            'orient': orient,
        }
        return meta.save()

    def link_swagger(self, specs, operations=None):
        """ link swagger specs to operations

        Args:
            specs (dict|filename): the swagger specifications
            operations (list|dict): the operations to support (names of services, models),
                if dict, a mapping of { '/api/path': 'name#action#method' }, where name is
                the name of a service or model, action is the method to call (e.g. predict),
                and method is the http method (e.g. post)

        Returns:
            list of mapped Metadata
        """
        specs = specs if isinstance(specs, dict) else markup(specs)
        assert specs.get('swagger') == '2.0', f"Only swagger version 2.0 is currently supported."
        paths = specs['paths']
        definitions = specs['definitions']
        actions = []
        operations = operations or self.list()
        metas = []

        def pathspecs():
            for path, pathspec in paths.items():
                for method, opspec in pathspec.items():
                    yield path, method, opspec

        for path, method, opspec in pathspecs():
            # operation mapping
            # -- format: name#action#method[,method]
            # -- name is the model or service name
            # -- action is the method to call (e.g. predict)
            # -- method(s) is the name of the http method
            opId = opspec.get('operationId')
            if opId is None:
                # extract from given operations spec
                if isinstance(operations, dict):
                    opId = operations.get(path)
                else:
                    # fallback to default operation
                    pathName = path.split('/')[-1]
                    opId = f'{pathName}#predict#post'
            name, action, methods = opId.split('#')
            if not self.exists(name):
                print(f"{name} not found in {self.prefix}, ignored.")
                continue
            methods = methods.split(',')
            actions.append((action, methods))
            # parameters
            params = opspec['parameters']
            for prmspec in params:
                if prmspec.get('in') != 'body':
                    continue
                # resolve schema spec
                # -- either direct spec of schema
                # -- or $ref: #/definitions/<name>, where name is the key into definitions
                schemaSpec = prmspec.get('schema')
                schemaRef = schemaSpec.get('$ref')
                Xschema = schemaSpec if schemaRef is None else definitions.get(schemaRef.split('/')[-1])
                break
            else:
                print("No body found, cannot generate X datatype")
                Xschema = None
            resps = opspec['responses']
            for code, respspec in resps.items():
                if code != '200':
                    continue
                schemaSpec = respspec.get('schema')
                schemaRef = schemaSpec.get('$ref')
                Yschema = schemaSpec if schemaRef is None else definitions.get(schemaRef.split('/')[-1])
                break
            else:
                print("No response 200 found, cannot generate Y datatype")
                Yschema = None
            # finally link
            Xdtype = self._datatype_from_schema(Xschema, name=f'{name}Input_{action}')
            Ydtype = self._datatype_from_schema(Yschema, name=f'{name}Output_{action}')
            meta = self.link_datatype(name, X=Xdtype, Y=Ydtype, actions=actions)
            metas.append(meta)
        return metas

    def validate(self, name, X=None, Y=None, result=None, **kwargs):
        meta = self.metadata(name, **kwargs)
        signature = meta.attributes.get('signature')

        def _do_validate(k, v):
            schema = (signature.get(k) or {}).get('schema')
            if schema:
                Schema = self._datatype_from_schema(schema, name=f'{name}_{k}')
                Schema().load(v)

        if signature:
            nop = lambda: ()
            _do_validate('X', X) if X is not None else nop()
            _do_validate('Y', Y) if Y is not None else nop()
            _do_validate('result', result) if result is not None else nop()
        return True

    def _datatype_from_schema(self, schema, name=None, orient='records'):
        # orient 'records' => every instance of the Schema reflects one item (default)
        #                     where each field is a distinct value of its type
        #        'columns' => every instance of the Schema reflects a full dataset, where
        #                     where each field is a list of values of its type
        # map OpenAPI 3.x to marshmallow.fields
        # -- nested types are not supported (except arrays)
        # https://swagger.io/specification/
        # https://marshmallow.readthedocs.io/en/stable/marshmallow.fields.html
        TYPE_MAP = {
            'integer': fields.Integer,
            'number': fields.Float,
            'string': fields.String,
            'string.byte': fields.String,
            'string.date-time': fields.DateTime,
            'string.date': fields.Date,
            'array': fields.List,
            'boolean': fields.Boolean,
            'password': fields.String,
            'default': fields.String,
            'object': fields.Raw,
        }
        prop = schema['properties']
        sdict = {}
        for prop, pspec in prop.items():
            ptype = pspec.get('type', 'object')
            pformat = pspec.get('format', '')
            fptype = '.'.join((ptype, pformat))
            ftype = TYPE_MAP.get(fptype) or TYPE_MAP.get(ptype) or TYPE_MAP.get('default')
            if ptype == 'array':
                ptype = TYPE_MAP.get(pspec['items'].get('type'))
                ftype = ftype(ptype) if ptype else ftype(fields.Dict())
            else:
                ftype = ftype()
            if orient == 'columns':
                ftype = fields.List(ftype)
            sdict[prop] = ftype
        schema = Schema.from_dict(sdict, name=name)
        schema.error_messages = {
            'type': f'invalid input for schema {name}'
        }
        return schema

    def _datatype_from_metadata(self, meta, orient='records'):
        # https://pbpython.com/pandas_dtypes.html
        TYPE_MAP = {
            "object": fields.String,
            "int": fields.Integer,
            "float": fields.Float,
            "dict": fields.Raw,
            "bool": fields.Boolean,
            "datetime": fields.DateTime,
            "timedelta[ns]": fields.DateTime,
            "date": fields.Date,
            "default": fields.String,
        }
        kind_meta = meta['kind_meta']
        dtypes = kind_meta.get('dtypes', {})
        name = meta.get('name')
        sdict = {}
        infer = lambda v: v.rstrip(string.digits)
        for col, colType in dtypes.items():
            if '#' in col:
                continue
            # orient == 'records' means every item is one value
            #        == 'columns' means every item is a list of values
            ftype = TYPE_MAP.get(infer(colType)) or TYPE_MAP.get('default')
            sdict[col] = ftype() if orient == 'records' else fields.List(ftype)
        schema = Schema.from_dict(sdict, name=name)
        schema.error_messages = {
            'type': f'invalid input for schema {name}'
        }
        return schema

    def _schema_from_datatype(self, datatype):
        spec = APISpec(
            title="omega-ml service",
            version="1.0.0",
            openapi_version="3.0",
            plugins=[MarshmallowPlugin()],
        )
        name = datatype.__name__
        spec.components.schema(name, schema=datatype)
        return spec.to_dict()['components']['schemas'][name]


class ScriptSignatureMixin(SignatureMixin):
    @classmethod
    def supports(cls, store, **kwargs):
        return store.prefix in ('scripts/')

    def _pre_run(self, scriptname, *args, **kwargs):
        validate_kwargs = {
            'X': args[0] if len(args) > 0 else kwargs.get('X'),
            'Y': args[1] if len(args) > 1 else kwargs.get('Y')
        }
        self.validate(scriptname, **validate_kwargs)
        return (scriptname, *args), kwargs

    def _post_run(self, result, scriptname, *args, om=None, **kwargs):
        meta = self.metadata(scriptname)
        self.validate(scriptname, result=result)
        return result


class ModelSignatureMixin(SignatureMixin):
    @classmethod
    def supports(cls, store, **kwargs):
        return store.prefix in ('models/')

    def link_dataset(self, name, Xname=None, Yname=None, Xmeta=None, Ymeta=None,
                     rName=None, features=None, labels=None, data_store=None,
                     meta=None, signature=True, actions=None, **kwargs):
        """ link dataset information to this model

        This sets the 'dataset' entry in metadata.attributes of a model. By default
        this method is called by any .fit() call initiated from the runtime. The 'dataset'
        entry is a dict that records the following information::

            {
                'Xname': Xname,  # the X dataset
                'Yname': Yname,  # the Y dataset
                'features': features, # features for this model
                'labels': labels, # labels for this model
                'Xmeta': Xmeta, # Metadata for this model (at time of fit)
                'Ymeta': Ymeta, # Metadata for this model (at time of fit)
                'rName': rName, # the result dataset
                'kwargs': kwargs, # other kwargs used for fitting
            }

        Args:
            name (str): the name of the model or script
            Xname (str): the name of dataset, may include modifiers
            Yname (str): the name of dataset, may include modifiers
            rName (str): the result dataset
            Xmeta (Metadata): the Metadata for X (features), defaults to data_store.metadata(Xname)
            Ymeta (Metadata): the Metadata for Y (features), defaults to data_store.metadata(Yname)
            features (list|dict): a list of feature names or dict(name=info), where info is a
               dict with further information for each features
            labels (list|dict): a list of label names or dict(label=info), where info is a
               dict with futher information for each label
            signature (bool): if True the object's signature will also be updated, defaults to True
            meta (Metadata): the metadata object, if provided .metadata() is not called again
            data_store (OmegaStore): the data store to retrieve Xmeta, Ymeta if not provided
            **kwargs: passed to self.metadata(name, **kwargs) to retrieve the model metadata
        """
        meta = meta or self.metadata(name, **kwargs)
        if data_store is None and (Xname or Yname):
            warnings.warn('Specify data_store=om.store to store metadata')
        metaX = (Xmeta or data_store.metadata(Xname)).to_mongo() if Xname and data_store else None
        metaY = (Ymeta or data_store.metadata(Yname)).to_mongo() if Yname and data_store else None
        model_attrs = {
            'Xname': Xname,
            'Yname': Yname,
            'features': features,
            'labels': labels,
            'Xmeta': metaX,
            'Ymeta': metaY,
            'rName': rName,
            'kwargs': kwargs,
        }
        meta.attributes.setdefault('dataset', {})
        meta.attributes['dataset'].update(model_attrs)
        if signature:
            self.link_datatype(name, X=self._datatype_from_metadata(metaX),
                               meta=meta, actions=actions) if metaX else None
            self.link_datatype(name, Y=self._datatype_from_metadata(metaY),
                               meta=meta, actions=actions) if metaY else None
        return meta.save()

    def link_dependencies(self, name, select=None, **kwargs):
        # TODO determine exact use and implementation
        meta = self.metadata(name, **kwargs)
        return meta.save()

    def _resolve_dataset_defaults(self, modelname, Xname, **kwargs):
        # use the linked dataset name if Xname == '*'
        kwargs = dict(kwargs)
        is_default = lambda name: name in ('*', None)

        def apply_default(key, kwargs):
            if key in kwargs and is_default(kwargs[key]):
                kwargs[key] = dataset_meta.get(key)

        meta = self.metadata(modelname)
        dataset_meta = meta.attributes.get('dataset', {})
        Xname = dataset_meta.get('Xname') if is_default(Xname) else Xname
        [apply_default(key, kwargs) for key in ('Yname', 'rName')]
        return (modelname, Xname), kwargs

    def _post_fit_link_dataset_(self, result, modelname, Xname, meta=None, data_store=None, **kwargs):
        meta = meta or self.metadata(modelname, **kwargs)
        self.link_dataset(modelname, Xname=Xname, meta=meta, data_store=data_store, **kwargs)
        return result

    def _pre_fit(self, modelname, Xname, **kwargs):
        # get default Xname, Yname if provided as '*'
        return self._resolve_dataset_defaults(modelname, Xname, **kwargs)

    def _pre_partial_fit(self, modelname, Xname, **kwargs):
        # get default Xname, Yname if provided as '*'
        return self._resolve_dataset_defaults(modelname, Xname, **kwargs)

    def _post_fit(self, result, modelname, Xname, **kwargs):
        # link the input dataset names with
        return self._post_fit_link_dataset_(result, modelname, Xname, **kwargs)

    def _post_partial_fit(self, result, modelname, Xname, **kwargs):
        return self._post_fit_link_dataset_(result, modelname, Xname, **kwargs)

    def _pre_predict(self, modelname, Xname, **kwargs):
        return self._resolve_dataset_defaults(modelname, Xname, **kwargs)

    def _pre_decision_function(self, modelname, Xname, **kwargs):
        return self._resolve_dataset_defaults(modelname, Xname, **kwargs)

    def _pre_predict_proba(self, modelname, Xname, **kwargs):
        return self._resolve_dataset_defaults(modelname, Xname, **kwargs)

    def _pre_score(self, modelname, Xname, **kwargs):
        return self._resolve_dataset_defaults(modelname, Xname, **kwargs)

    def _pre_transform(self, modelname, Xname, **kwargs):
        return self._resolve_dataset_defaults(modelname, Xname, **kwargs)

    def _pre_fit_transform(self, modelname, Xname, **kwargs):
        return self._resolve_dataset_defaults(modelname, Xname, **kwargs)



import re
from importlib import import_module

from pathlib import Path

def migrate(om):
    for modfile in Path(__file__).parent.glob('*.py'):
        modname = modfile.name.replace('.py', '')
        if not re.match(r'^[0-9]{4}_.*', modname):
            continue
        print(f"running migration {modname}")
        migration = import_module('.' + modname, package=__package__)
        migration.forward(om)
    print(f"Migrations complete.")



import unittest

import pandas as pd

from omegaml import Omega
from omegaml.backends.experiment import ExperimentBackend
from omegaml.tests.util import OmegaTestMixin
from omegaml.util import module_available


@unittest.skipUnless(module_available("tensorflow"), "tensorflow is not installed")
class TFCallbackTrackingTestCases(OmegaTestMixin, unittest.TestCase):
    """
    notes on framework versions v.v. tracing

    tensorflow
    1.15.3  - does not report metric on epoch, while later versions do
              (=> reports only 9 instead of 10 metrics per run)
            - reports metric='accuracy' as 'acc'
    """

    def setUp(self):
        self.om = om = Omega()
        self.clean()
        om.models.register_backend(ExperimentBackend.KIND, ExperimentBackend)

    def test_tensorflow_callback(self):
        om = self.om
        # fit locally
        with om.runtime.experiment('myexp') as exp:
            model, X, Y = self._create_model(exp.tensorflow_callback())
        self.assertIsInstance(exp.data(), pd.DataFrame)
        self.assertGreaterEqual(len(exp.data(key='loss')), 9)
        self.assertGreaterEqual(len(exp.data(key=['acc', 'accuracy'])), 9)
        model_ = exp.restore_artifact('model')
        self.assertIsInstance(model, type(model_))
        # fit via runtime
        om.models.put(model, 'mymodel')
        with om.runtime.experiment('myexp2') as exp:
            om.runtime.model('mymodel').fit(X, Y, epochs=1,
                                            batch_size=128).get()
        self.assertIsNotNone(exp.data())
        self.assertGreaterEqual(len(exp.data(key=['acc', 'accuracy'])), 9)

    def _create_model(self, tracking_cb):
        import numpy as np
        import tensorflow as tf
        from tensorflow import keras
        from tensorflow.keras.optimizers import SGD
        from tensorflow.keras.models import Sequential
        from tensorflow.keras.layers import Dense

        x_train = np.random.random((1000, 20))
        y_train = keras.utils.to_categorical(np.random.randint(10, size=(1000, 1)), num_classes=10)

        model = Sequential()
        # FIXME tv versions 2.2 through 2.6 worked with x_train.shape, 1.15.3 and since 2.7 dont
        if tf.version.VERSION == '1.15.3':
            # rationale: https://stackoverflow.com/a/43233458/890242
            x_shape = x_train.shape[1:]
        else:
            x_shape = x_train.shape
        x_shape = x_train.shape[1:]
        model.add(Dense(10, activation='softmax', input_shape=x_shape))
        sgd = SGD(lr=0.01, decay=1e-6, momentum=0.9, nesterov=True)
        model.compile(loss='categorical_crossentropy',
                      optimizer=sgd,
                      metrics=['accuracy'])
        model.fit(x_train, y_train,
                  epochs=1,
                  batch_size=128, callbacks=[tracking_cb])
        return model, x_train, y_train


if __name__ == '__main__':
    unittest.main()



from requests.auth import AuthBase

from omegaml import session_cache
from omegaml.util import load_class, settings


class OmegaRestApiAuth(AuthBase):
    """
    Sets the appropriate authentication headers
    for the Omega REST API key authentication.

    Usage:
        auth = OmegaRestApiAuth('jezdez',
                         '25fdd0d9d210acb78b5b845fe8284a3c93630252')
        response = requests.get('http://api.foo.bar/v1/spam/', auth=auth)
    """

    def __init__(self, username, apikey, qualifier=None):
        self.username = username
        self.apikey = apikey
        self.qualifier = qualifier or 'default'

    def get_credentials(self):
        if self.kind == 'jwt':
            return 'Bearer %s' % (self.apikey,)
        return 'ApiKey %s:%s' % (self.username, self.apikey)

    def __call__(self, r):
        r.headers['Authorization'] = self.get_credentials()
        r.headers['Qualifier'] = self.qualifier
        return r

    def __repr__(self):
        return ('OmegaRestApiAuth(username={}, apikey="*****",'
                'qualifier={})').format(self.username, self.qualifier)

    @property
    def kind(self):
        return 'jwt' if self.username.startswith('jwt:') or self.apikey.count('.') == 2 else 'apikey'


class OmegaRuntimeAuthentication:
    """
    The runtimes authentication
    """

    def __init__(self, userid, apikey, qualifier='default'):
        self.userid = userid
        self.apikey = apikey
        self.qualifier = qualifier

    @property
    def token(self):
        return self.userid, self.apikey, self.qualifier

    def __repr__(self):
        return ('OmegaRuntimeAuthentication(userid={}, '
                'apikey="*****", qualifier={})').format(self.userid, self.qualifier)


class AuthenticationEnv(object):
    """ AuthenticationEnv creates Omega() configured to the current environment

    This provides a basic authentication environment that is not protected. It
    should not be used in production environments.

    Usage::

        auth_env = AuthenticationEnv.active()
        om = auth_env.get_omega_from_apikey(<credentials>)
        om = auth_env.get_omega_for_task(task, auth=<credentials>)

    Notes:
        * AuthenticationEnv.active() returns the environment as configured
          in defaults.OMEGA_AUTH_ENV. This defaults to AuthenticationEnv.
        * Any AuthenticationEnv is required to provide at least
          get_omega_from_apikey() and get_omega_from_task() methods, as
          the primary ways to instantiate omega is from either a set of
          credentials (userid, apikey), or from a an authorized task that
          provides these credentials
        * some authentication environments may require additional methods
          to provide credentials in a format suitable for their authentication
          backends (e.g. REST APIs, kerberos, etc.)
    """
    auth_env = None
    is_secure = False

    @classmethod
    @session_cache  # PERFTUNED
    def get_omega_for_task(cls, task, auth=None):
        # return the omega instance for the given task authentication
        from omegaml import setup
        om = setup()
        return om

    @classmethod
    @session_cache  # PERFTUNED
    def get_omega_from_apikey(cls, auth=None):
        # return the omega instance for the given task authentication
        from omegaml import setup
        om = setup()
        return om

    @classmethod
    def active(cls):
        # load the currently active auth env
        if cls.auth_env is None:
            from omegaml import _base_config
            cls.auth_env = load_class(getattr(_base_config, 'OMEGA_AUTH_ENV',
                                              'omegaml.client.auth.AuthenticationEnv'))
        return cls.auth_env

    @classmethod
    def secure(cls):
        # load a server-backed authentication env
        from omegaml import _base_config
        if not hasattr(_base_config, 'OMEGA_AUTH_ENV'):
            _base_config.OMEGA_AUTH_ENV = 'omegaml.client.auth.CloudClientAuthenticationEnv'
            cls.auth_env = None
        auth_env = cls.active()
        if not auth_env.is_secure:
            raise SystemError(f'A secure authentication environment was requested, however {auth_env} is not secure.')
        return auth_env

    @classmethod
    def taskauth(cls, args, kwargs, celery_kwargs):
        # apply event mask in monitoring events
        nop = lambda *args, **kwargs: ()
        getattr(cls, 'taskrepr', nop)(args, kwargs, celery_kwargs)

    @classmethod
    def resultauth(cls, value):
        # apply result mask in monitoring events
        nop = lambda v: value
        getattr(cls, 'resultrepr', nop)(value)


class CloudClientAuthenticationEnv(AuthenticationEnv):
    is_secure = True

    @classmethod
    @session_cache
    def get_omega_for_task(cls, task, auth=None):
        """
        Get Omega instance configured for user in auth

        If auth is passed, a request is made to OMEGA_RESTAPI_URL to
        retrieve the configuration object for this user.

        If auth is the tuple (None, None, 'default') the omegaml module
        is returned, which is configured to the default instance with
        authentication according to the installation. To raise an
        error instead set settings OMEGA_ALLOW_TASK_DEFAULT_AUTH=False

        :param auth: the OmegaRuntimeAuthentication object, or a token (userid, apikey, qualifier)
        :return: the Omega instance configured for the user
        """
        from omegaml.util import settings

        default_auth = (None, None, 'default')
        is_auth_provided = lambda token: (token is not None
                                          and token != default_auth)
        defaults = settings()
        token = auth.token if isinstance(auth, OmegaRuntimeAuthentication) else auth

        if is_auth_provided(token):
            if isinstance(token, (list, tuple)):
                # we get a serialized tuple, recreate auth object
                userid, apikey, qualifier = token
                # by default assume worker is in cluster
                # TODO refactor this setting to eedefaults
                view = defaults.OMEGA_SERVICES_INCLUSTER
                om = cls.get_omega_from_apikey(userid, apikey, qualifier=qualifier, view=view)
            else:
                raise ValueError(
                    'cannot parse authentication as {}'.format(auth))
        elif token == default_auth:
            # we provide the default implementation as per configuration
            from omegaml import _omega
            om = _omega._om
            if not getattr(defaults, 'OMEGA_ALLOW_TASK_DEFAULT_AUTH', True):
                raise ValueError(
                    'Default task authentication is not allowed, got {}'.format(auth))
        else:
            raise ValueError(
                'Missing runtime task authentication, got {}'.format(auth))
        return om

    @classmethod
    @session_cache  # PERFTUNED
    def get_omega_from_apikey(cls, *args, **kwargs):
        from omegaml.client.userconf import _get_omega_from_apikey
        return _get_omega_from_apikey(*args, **kwargs)

    @classmethod
    def get_restapi_auth(cls, defaults=None, om=None,
                         userid=None, apikey=None, qualifier=None):
        assert defaults or om, "require either defaults or om"
        defaults = defaults or om.defaults
        return OmegaRestApiAuth(userid or defaults.OMEGA_USERID,
                                apikey or defaults.OMEGA_APIKEY,
                                qualifier or defaults.OMEGA_QUALIFIER)

    @classmethod
    def get_runtime_auth(cls, defaults=None, om=None):
        assert defaults or om, "require either defaults or om"
        defaults = defaults or om.defaults
        return OmegaRuntimeAuthentication(defaults.OMEGA_USERID,
                                          defaults.OMEGA_APIKEY,
                                          defaults.OMEGA_QUALIFIER)

    @classmethod
    def get_userconfig_from_api(cls, api_auth=None, api_url=None, userid=None, apikey=None,
                                requested_userid=None, defaults=None, qualifier=None, view=False):
        from omegaml.client.userconf import _get_userconfig_from_api, ensure_api_url
        defaults = defaults or settings()
        api_auth = api_auth or cls.get_restapi_auth(userid=userid, apikey=apikey,
                                                    qualifier=qualifier,
                                                    defaults=defaults)
        return _get_userconfig_from_api(api_auth,
                                        api_url=ensure_api_url(api_url, defaults),
                                        qualifier=qualifier or defaults.OMEGA_QUALIFIER,
                                        requested_userid=requested_userid,
                                        view=view)

    @classmethod
    def save_userconfig_from_apikey(cls, configfile, userid, apikey, api_url=None, requested_userid=None,
                                    view=False, keys=None, qualifier=None):
        from omegaml.client.userconf import _save_userconfig_from_apikey
        return _save_userconfig_from_apikey(configfile, userid, apikey, api_url=api_url,
                                            requested_userid=requested_userid,
                                            view=view, keys=keys, qualifier=qualifier)

    @classmethod
    def taskrepr(cls, args, kwargs, celery_kwargs):
        celery_kwargs.update({
            'argsrepr': ['*****'] * len(args),
            'kwargsrepr': {k: '*****' for k in kwargs}
        })

    @classmethod
    def resultrepr(cls, value):
        return value



class GridSearchMixin(object):
    def gridsearch(self, Xname, Yname=None, parameters=None, pure_python=False, **kwargs):
        """ run gridsearch on model

        Args:
            Xname (str|obj): the name of the X dataset in om.datasets, or
                the data object
            Yname (str|obj): the name of the Y dataset in om.datasets, or
                the data object
            parameters (dict): input to GridSearchCV(..., param_grid=parameters)

        See Also:
            * sklearn.model_selection.GridSearchCV
        """
        gs_task = self.task('omegaml.tasks.omega_gridsearch')
        Xname = self._ensure_data_is_stored(Xname, prefix='_fitX')
        if Yname is not None:
            Yname = self._ensure_data_is_stored(Yname, prefix='_fitY')
        return gs_task.delay(self.modelname, Xname, Yname=Yname, parameters=parameters, **kwargs)



from unittest import TestCase

from omegaml import Omega
from omegaml.tests.util import OmegaTestMixin


class StreamsTestCase(OmegaTestMixin, TestCase):
    # tests om.streams, see minibatch functionality in test_minibatch
    def setUp(self):
        self.om = Omega()
        self.clean()

    def test_streams_get_put(self):
        om = self.om
        stream = om.streams.get('test')
        om.streams.put(dict(foo='bar'), 'test')
        self.assertEqual(stream.buffer().count(), 1)
        meta = om.streams.metadata('test')
        self.assertEqual(meta.kind, 'stream.minibatch')

    def test_streams_emitter(self):
        from minibatch.window import CountWindow
        from minibatch.tests.util import LocalExecutor
        om = self.om
        runner = om.streams.getl('test', size=1, executor=LocalExecutor())
        emitter = runner.make(lambda w: w)
        self.assertIsInstance(emitter, CountWindow)
        om.streams.put(dict(foo='bar'), 'test')
        self.assertEqual(om.streams.get('test').buffer().count(), 1)
        emitter.run(blocking=False)
        self.assertEqual(om.streams.get('test').buffer().count(), 0)







from omegaml.documents import MDREGISTRY


class PromotionMixin(object):
    """
    Promote objects from one bucket to another
    """
    DEFAULT_DROP = {
        'models/': False, # models are versioned, don't drop
    }

    @classmethod
    def supports(cls, store, **kwargs):
        return store.prefix in ('data/', 'jobs/', 'models/', 'scripts/', 'streams/')

    def promote(self, name, other, asname=None, drop=None, **kwargs):
        """
        Promote object to another bucket.

        This effectively copies the object. If the objects exists in the
        target it will be replaced.

        Args:
            name: The name of the object
            other: the OmegaStore instance to promote to
            asname: the name to use in other, defaults to .metadata(name).name
            kwargs: will be forwarded to other.put

        Returns:
            The Metadata of the new object
        """
        drop = drop if drop is not None else self.DEFAULT_DROP.get(self.prefix, True)
        # sanity checks
        # -- promotion by same name requires a different store
        meta = self.metadata(name)
        asname = asname or meta.name
        if name == asname and self == other:
            raise ValueError(f'must specify asname= different from {meta.name}')
        # see if the backend supports explicit promotion
        backend = self.get_backend(name)
        if hasattr(backend, 'promote'):
            return backend.promote(name, other, asname=asname, **kwargs)
        # do default promotion, i.e. copy
        meta = self.metadata(name)
        obj = self.get(name)
        asname = asname or meta.name
        other.drop(asname, force=True) if drop else None
        # TODO refactor to promote of python native data backend
        if meta.kind == MDREGISTRY.PYTHON_DATA:
            # in case of native python data we get back a list of
            # all previously inserted objects. do the same in other
            [other.put(o, asname) for o in obj]
            other_meta = other.metadata(asname)
        else:
            other_meta = other.put(obj, asname, **kwargs)
        return other_meta



from unittest import TestCase

from dashserve.jupyter import DashserveApp
from dashserve.tests import myapp
from omegaml.backends.dashapp import DashAppBackend
from omegaml.store import OmegaStore


class DashAppBackendTests(TestCase):
    def setUp(self):
        self.app = myapp.app
        self.store = OmegaStore(prefix='scripts/')
        self.backend = DashAppBackend(model_store=self.store, data_store=self.store)
        [self.store.drop(name) for name in self.store.list()]

    def tearDown(self):
        pass

    def test_supports(self):
        """ test DashAppBackend supports a jupyter dash app """
        self.assertTrue(DashAppBackend.supports(self.app, 'myapp'))

    def test_put_get_dashapp(self):
        """ test a jupyter dash app can be stored and restored"""
        app = self.app
        self.backend.put(self.app, 'myapp')
        self.assertIn('myapp', self.store.list())
        restored_app = self.backend.get('myapp')
        self.assertIsInstance(restored_app, DashserveApp)
        self.assertEqual(restored_app.config.external_stylesheets , app.config.external_stylesheets)









def forward(om):
    for meta in om.models.list(raw=True):
        if 'dataset' in meta.attributes and isinstance(meta.attributes.get('dataset'), str):
            meta.attributes.setdefault('tracking', {})
            meta.attributes['tracking']['dataset'] = meta.attributes['dataset']
            del meta.attributes['dataset']
            meta.save()



import smart_open

from omegaml.client.cli.stores import StoresCommandMixin
from omegaml.client.docoptparser import CommandBase
from omegaml.client.util import get_omega


class DatasetsCommandBase(StoresCommandMixin, CommandBase):
    """
    Usage:
      om datasets list [<pattern>] [--raw] [-E|--regexp] [options]
      om datasets put <path> <name> [--replace] [--csv=<param=value>]... [--format csv|image|binary] [options]
      om datasets get <name> <path> [--csv <param>=<value>]... [--format csv|binary] [options]
      om datasets drop <name> [--force] [options]
      om datasets metadata <name> [options]

    Options:
      --raw           return metadata
      --format VALUE  force the format of the file

    Description:
         For csv files, put and get accept the --csv option multiple times.
         The <param>=<value> pairs will be used as kwargs to pd.read_csv (on put)
         and df.to_csv methods (on get). Note that put uses read_csv because
         it must read the csv first, then store in om.datasets. Respectively,
         get uses to_csv because it must store the file locally.
    """
    command = 'datasets'

    def put(self):
        om = get_omega(self.args)
        local = self.args['<path>']
        name = self.args['<name>']
        replace = self.args['--replace']
        csvkwargs = self.parse_kwargs('--csv')
        # TODO introduce a puggable filetype processing backend to do this
        is_csv = self.args.get('--format') == 'csv' or anyext(local, '.csv')
        is_image = self.args.get('--format') == 'image' or anyext(local,
                                                                  '.png,.img,.bmp,.jpeg,.jpg,.tif,.tiff,.eps,.raw,.gif')
        is_binary = self.args.get('--format') == 'binary' or not (is_csv or is_image)
        if is_csv:
            # csv formats
            om.datasets.read_csv(local, name, append=not replace, **csvkwargs)
            meta = om.datasets.metadata(name)
        elif is_image:
            # images
            from imageio import imread
            with smart_open.open(local, 'rb') as fin:
                img = imread(fin)
                meta = om.datasets.put(img, name)
        elif is_binary:
            with smart_open.open(local, 'rb') as fin:
                meta = om.datasets.put(fin, name, append=not replace)
        else:
            meta = om.datasets.put(local, name, append=not replace)
        self.logger.info(meta)

    def get(self):
        om = get_omega(self.args)
        local = self.args['<path>']
        name = self.args['<name>']
        try:
            # try lazy get first to suppport large dataframes
            obj = om.datasets.get(name, lazy=True)
        except:
            obj = om.datasets.get(name)
        csvkwargs = self.parse_kwargs('--csv', index=False)
        is_csv = self.args.get('--format') == 'csv' or hasattr(obj, 'to_csv') or anyext(local, '.csv')
        is_binary = self.args.get('--format') == 'binary' or hasattr(obj, 'read')
        if is_csv and not is_binary:
            obj.to_csv(local, **csvkwargs)
        elif is_binary:
            with smart_open.open(local, 'wb') as fout:
                while True:
                    data = obj.read(1024 * 10)
                    if not data:
                        break
                    fout.write(data)
        self.logger.debug(local)

anyext = lambda n, exts: any(n.endswith(ext) for ext in exts.split(','))



"""
implements REST API resource helpers to process async results (promises)

AsyncResponseMixin - a mixin to enable a given Resource to return actual and async results, from the same code
AsyncTaskResourceMixin - a mixin to enable the TaskResource to return status information and resolve results given a task id
"""
from http import HTTPStatus

import celery
import flask
from celery.result import AsyncResult, EagerResult
from flask import request
from werkzeug.exceptions import NotFound

EAGER_RESULTS = {}

truefalse = lambda v: (v if isinstance(v, bool) else
                       any(str(v).lower().startswith(c) for c in ('y', 't', '1')))


class AsyncResponseMixin:
    """
    Mixin to enable a Flask RESTPlus API Resource to return sync and async results with a simple modification

    This implements the async REST API protocol as per http://restcookbook.com/Resources/asynchroneous-operations/
    Specifically for async responses, it returns

        HTTP 202/ACCEPTED
        Location: /location/to/resolve/result

    The same Resource and code can be used to return sync and async results, where the user can request the async version
    by providing the ASYNC=true header, or the ?async=true query argument

    Usage:

        @api.route(...)
        class SomeResource(AsyncResponseMixin, Resource):
            def get(...):
                result = ... # dict or AsyncResult
                async_body = { ... } # dict of data to include if result is AsyncResult
                return self.maybe_async(result, async_body={...})  # would usually return (result, status, headers)

        self.maybe_async() returns, if result is

        * not of type AsyncResult: (body, status, headers), as any usual Resource method
        * of type AsyncResult: (body, status, headers), where
            - status = HTTPStatus.ACCEPTED (202)
            - headers contains a Location header to resolve the result as per self.result_uri
            - a body referencing the AsyncResult task id and the original resources route
              {
                'resource_uri': self.resource_uri,
                'task_id': async.id,
              }

        Clients can request async or sync responses by specifying either of

        * request header ASYNC: true
        * url request query ?async=true

        Your resource method should check the self.is_async property to determine if it should return sync or async
        results. Regardless, self.maybe_async() determines its response by the actual result, not the self.is_async
        value.

        In some cases you may want to process the deferred result returned by the task to return a format in line
        with your REST API specification. A common way to implement task result processing is as follows:

            class SomeResource(...):
                def get(...):
                    async_result = ...
                    async_body = { ... } # dict of data to include if result is AsyncResult
                    result = self.prepare_result(async_result.get()) if not self.is_async else async_result
                    return self.maybe_async(result, async_body=async_body)

                def prepare_result(self, value, **kwargs):
                    return { ... } # actual value

        Note the only difference to the first example is the prepare_result() call and method. The reason for having
        prepare_result() as a separate method is so that it can be reused in the TaskResource to process the deferred
        result value of the AsyncResult. This ensures that the sync and async responses are equivalent, simplifying
        the REST API client processing.

        See the AsyncTaskResourceMixin for details on status checking and deferred result resolving.

    Configuration:

        class SomeResource(AsyncResponseMixin, Resource):
            # specify the client reachable URI to resolve the task, {id} is replaced with task.id
            result_uri = '/api/task/{id}/result'

    """
    # Flask Restplus Resource adapter
    result_uri = '/task/{id}/result'

    def dispatch_request(self, *args, **kwargs):
        # inline with we do not use x-async https://tools.ietf.org/html/rfc6648
        # note in Flask request headers are always as-is and lower-case
        # see https://stackoverflow.com/a/57562733/890242
        self.is_async = (truefalse(request.args.get('async', False)) or
                         truefalse(request.headers.get('async', False)))
        return super().dispatch_request(*args, **kwargs)

    @property
    def resource_uri(self):
        return flask.request.path

    def prep_response(self, result, status=None, headers=None, async_body=None, **kwargs):
        if isinstance(result, AsyncResult):
            if isinstance(result, EagerResult):
                EAGER_RESULTS[result.id] = result
            headers = headers or {}
            headers.update({
                'Location': self.result_uri.format(id=result.id)
            })
            body = async_body or {}
            body.update({
                'resource_uri': self.resource_uri,
                'task_id': result.id,
            })
            status = status or HTTPStatus.ACCEPTED
        else:
            body, status, headers = result, status or HTTPStatus.OK, {}
        return body, int(status), headers

    maybe_async = prep_response
    create_maybe_async_response = prep_response


class AsyncResponseMixinTastypie(AsyncResponseMixin):
    """
    AsyncResourceMixin for Django Tastypie Resources

    Usage:

        class SomeResource(AsyncResponseMixinTastypie, tastypie.Resource):
            class Meta:
                ...
                result_uri = '/api/task/{id}/result'

            def get_detail(self, request, **kwargs):
                result = ... # dict or AsyncResult
                async_body = { ... } # dict of data to include if result is AsyncResult
                # would usually return create_response(...)
                return self.create_maybe_async_response(request, result, async_body={...})

        class TaskResource(AsyncTaskResourceMixin, tastypie.Resource):
            ...
            def get_detail(self, request, **kwargs):
                taskid = kwargs.get('pk') # <pk> is specified in uri by default
                action = kwargs.get('action') # assumes <action> is specified in uri, see Resource.prepend_url
                value, status = self.process_task_action(taskid, action, request=request)
                resp = self.create_response(request, value, status=status)
                return resp
    """

    def dispatch(self, request_type, request, *args, **kwargs):
        # inline with we do not use x-async https://tools.ietf.org/html/rfc6648
        # note in Django request headers are always prefixed with HTTP, uppercase
        # see https://docs.djangoproject.com/en/3.0/ref/request-response/#django.http.QueryDict
        self.is_async = (truefalse(request.GET.get('async', False)) or
                         truefalse(request.META.get('HTTP_ASYNC', False)))
        self._resource_uri = request.path
        return super().dispatch(request_type, request, **kwargs)

    def create_maybe_async_response(self, request, result, status=None, async_body=None):
        # tastypie.create_response drop-in to match AsyncResponseMixin.maybe_async compatibility
        body, status, headers = self.maybe_async(result, status=status, async_body=async_body)
        resp = self.create_response(request, body, status=int(status))
        for k, v in headers.items():
            resp[k] = v
        return resp

    @property
    def result_uri(self):
        return self._meta.result_uri

    @property
    def resource_uri(self):
        return self._resource_uri


class AsyncTaskResourceMixin:
    """
    Mixin to simplify implementation of a TaskResource that returns async results

    Usage:
        To support /task/<id>/status and /task/<id>/result actions implement a resource as follows:

            @api.route('/api/v1/task/<string:taskid>/<string:action>')
            class TaskResource(AsyncTaskResourceMixin, Resource):
                def get(self, taskid, action):
                    # do custom processing or initialization here if needed
                    # any kwargs are passed along as the context argument in subsequent calls
                    return self.process_task_action(taskid, action, request=flask.request)

        With this, the AsyncTaskResourceMixin will take care of /status and /result actions. In particular:

        * /status => self.get_task_status():
            default implementation returns
            {
              'task_id': taskid,
              'status': promise.status,
              'message': message,
            }

        * /result => self.get_task_result():
            default implementation returns
            {
                'task_id': taskid,
                'status': status,
                'response': data
            }

        * where the /result {response: data } is resolved by calling

                promise = self.get_async_result(taskid, context)
                value = self.resolve_promise(promise, context) # calls promise.get()
                resource_method, method_kwargs = self.resolve_resource_method(taskid, context)
                data = resource_method(value, **method_kwargs)

            The default resource_method will simply return the value as is, implemented in self.default_prepare_result()

        * To apply custom value processing, e.g. to support different types of resources or to colocate prep logic with
          the original Resource (instead of duplicating in TaskResource), implement the following additional method:

            def resolve_resource_method(...):
                # return the (callable, kwargs) tuple to format the task results for the actual response
                # the callable should be the same that prepares the result for your sync API response, e.g. prepare_result
                # res_kwargs will be passed on as resource.prepare_result(value, **res_kwargs)
                resourceUri = api.payload.get('resource_uri')
                path, res_kwargs = resolve(resourceUri) # path is the endpoint (str) that matches resource_uri, res_kwargs are the matched kwargs of the route
                resource = <determine resource from path>
                return resource.prepare_result, res_kwargs

          Here the 'resource_uri' is provided in the request payload, referencing the original resource route used to generate
          the async result (i.e. by your SomeResource). The provided resource method will be called as method(value, **kwargs).
    """
    celeryapp = celery.current_app
    is_async = False  # AsyncTaskResource does not support async processing by itself

    def process_task_action(self, taskid, action, **context):
        """
        given a task id and action call the respective get_task_<action> method

        This calls the get_task_<action> method, passing along the caller_args and
        caller_kwargs. The latter should contain any information subsequent

        Args:
            taskid:
            action:
            *caller_args:
            **caller_kwargs:

        Returns:

        """
        promise = self.get_async_result(taskid, context)
        action_meth = getattr(self, 'get_task_{}'.format(action), None)
        if action_meth is None:
            raise ValueError('unknown action {} on task {}'.format(action, taskid))
        try:
            value = action_meth(promise, taskid, context)
            status = HTTPStatus.OK
        except Exception as e:
            value = self.get_task_status(promise, taskid, context, message=str(e))
            status = HTTPStatus.BAD_REQUEST
        return value, int(status)

    def get_task_status(self, promise, taskid, context, message=None):
        result = {
            'task_id': taskid,
            'status': promise.status,
            'message': message,
        }
        return result

    def get_task_result(self, promise, taskid, context):
        # resolve the async result and
        status = promise.status
        value = self.resolve_promise(promise, context)
        result_method, kwargs = self.resolve_resource_method(taskid, context)
        data = result_method(value, **kwargs)
        result = {
            'task_id': taskid,
            'status': status,
            'response': data
        }
        return result

    def get_async_result(self, task_id, context):
        # from a given task id return a AsyncResource as promise
        # hack to allow local testing
        if not getattr(self.celeryapp.conf, 'CELERY_ALWAYS_EAGER'):
            promise = self.celeryapp.AsyncResult(task_id)
        else:
            promise = EAGER_RESULTS[task_id]
        return promise

    def resolve_promise(self, promise, context):
        # resolve a given AsyncResult, as returned by get_async_result
        if promise.ready():
            try:
                value = promise.get()
            except Exception as error:
                raise ValueError(error)
        else:
            value = None
        return value

    def resolve_resource_method(self, taskid, context):
        # resolve the GenericResoure method to prepare the final result
        # return (callable, kwargs)
        return self.default_prepare_result, {'context': context}

    def default_prepare_result(self, value, **kwargs):
        return value


def resolve(uri, method='GET,PUT,POST,DELETE,PATCH'):
    # Flask equivalent of django.urls.resolve
    for m in method.split(','):
        try:
            # see https://werkzeug.palletsprojects.com/en/1.0.x/routing/#werkzeug.routing.MapAdapter.match
            result = flask.current_app.url_map.bind('localhost').match(uri, method=m)
        except:
            pass
        else:
            return result
    raise NotFound("path {} not found".format(uri))



import callable_pip as cpip
import logging
import os
import sys
from pathlib import Path

logger = logging.getLogger(__name__)


def build_sdist(src, distdir):
    # TODO replace with python -m build
    #      https://setuptools.pypa.io/en/latest/userguide/quickstart.html
    from setuptools.sandbox import run_setup
    if not src.endswith('setup.py'):
        setup_path = os.path.join(src, 'setup.py')
    else:
        setup_path = src
    run_setup(setup_path, ['sdist', '--dist-dir', distdir])
    sdist = sorted(Path(distdir).iterdir(), key=os.path.getmtime)[-1]
    return str(sdist)


def install_package(src, dst):
    cpip.main('install',
              src,
              '--force-reinstall',
              '--no-cache-dir',
              '--upgrade',
              '--target', dst)


def load_from_path(name, path, keep=False):
    import importlib
    sys.path.insert(0, path)
    try:
        if name in sys.modules:
            del sys.modules[name]
        if name in globals():
            del globals()[name]
        mod = importlib.import_module(name)
    finally:
        # remove the import path -- this is to avoid
        # however comes at the cust
        if not keep:
            sys.path.remove(path)
    return mod


def install_and_import(pkgfilename, package, installdir, keep=False):
    """

    Returns:
        object:
    """
    install_package(pkgfilename, installdir)
    mod = load_from_path(package, installdir, keep=keep)
    return mod


class RunnablePackageMixin:
    # mixin to Package backends to enable backend.perform('run') from tasks
    def run(self, name, *args, keep=False, om=None, **kwargs):
        mod = self.get(name, keep=keep)
        return mod.run(om, *args, **kwargs)


if __name__ == '__main__':
    basepath = os.path.join(os.path.dirname(__file__), '..')
    pkgpath = os.path.join(basepath, 'demo/helloworld')
    build_sdist(pkgpath)

    pkgfilename = os.path.join(basepath, 'demo/helloworld/dist/helloworld-1.0.tar.gz')
    installdir = '/tmp/xyz'
    install_and_import(pkgfilename, 'helloworld', installdir)

    import helloworld

    helloworld.hello()



from unittest import TestSuite, TestLoader, TextTestRunner

# list all test cases to be called from within reticulated python
from omegaml.runtimes.rsystem import rhelper
from omegaml.util import temp_filename, remove_temp_filename

RE_TEST_CASES = ['omegaml.tests.rsystem.test_rmodels.RSystemModelTests',
                 'omegaml.tests.rsystem.test_rscripts.RSystemScriptTests']

def R_unittests():
    # -- build test suite, adopted from https://docs.python.org/3/library/unittest.html#load-tests-protocol
    suite = TestSuite()
    loader = TestLoader()
    runner = TextTestRunner()
    tests = loader.loadTestsFromNames(RE_TEST_CASES)
    suite.addTests(tests)
    # -- run the suite
    #    this will print any failures in stderr
    runner.run(suite)


def r_source(r_code):
    """ helper to source given r code and reticulated r session

    Args:
            r_code (str): the string of R code, as entered in REPL or R file
    """
    r = rhelper()
    fn = temp_filename(ext='R')
    with open(fn, 'w') as fout:
        fout.write(r_code)
    result = r.source(fn)
    remove_temp_filename(fn)
    return result



import pandas as pd

# code to insert into first and last cell of generated jobs
init_cell_code = """
## generated by omegaml-nbtasks
if '{userid}' != '':
    from omegaml.client.auth import AuthenticationEnv
    auth_env = AuthenticationEnv.secure()
    om = auth_env.get_omega_from_apikey(userid='{userid}', apikey='{apikey}')
else:
    import omegaml as om 
job = {job}
def mark(status):
    task_name = job.get('task_name', '__testing__')
    if task_name == '__testing__':
        print("warning: test run, not recording status")
    meta = om.jobs.metadata(task_name)
    job_meta = meta.attributes['job']
    job_meta['status'] = status
    meta.save()
mark('pending')
"""
done_cell_code = """
## generated by omegaml-nbtasks
mark('finished')
"""


class JobTasks:
    """
    example notebook task runner using omegaml runtime

    Parallel execution of notebooks with parameters on the omegaml runtime

    Usage::

      # submit tasks
      # -- creates 10 tasks as copies of the main notebook in om.jobs.list('tasks')
      # -- runs each task using omegaml runtime
      # -- every .map() call generates a new task group, ensuring unique notebook names
      # -- the original notebook is not changed
      job = om.runtime.job('main')
      job.map(range(10))

      # check status
      job.status()

      # restart tasks that did not produce a result yet
      job.restart()

      # get the list of all notebooks created in one .map() call
      job.list()

      # get the list of all notebooks in any .map() call
      job.list(task_group='*')

      # example notebook
      job = globals().get('job', dict(param=my_default_value))

      def calculation(job):
         # job is a dict with job['param'] set to one value taken from .map(<jobs>)
         # job contains other keys to identify the task:
         #   job_id: a sequence number
         #   task_group: the group of tasks submitted in one .map() call
         #   task_name: the om.jobs name of the running notebook
         #   status: the status, pending or finished
         #   task_id: the celery task id, if available
         ... <insert your code here>

      # run the calculation
      # -- in each task, job will be one of the variables given in nbtasks.map(<jobs>)
      #    <jobs> is an iterable, returning one object for each job
      #    note the job must be serializable. if you need something more complex, pass
      #    the name of an omegaml dataset and possibly some query criteria
      calculation(job)
    """

    def _init_mixin(self, *args, **kwargs):
        self.task_group = None

    def map(self, jobs, job_ids=None, require=None, reset=False, task_group=None):
        """
        Generate any number of parallel jobs executed through om.runtime

        Args:
            jobs (iterable): an iterable to yield job parameters, each parameter
              can be any native object or container (e.g. int, float, dict, list)
            job_ids (str): optional, list of job ids. If passed, the
               job id will be used to name the task id, else
               it is the current count
            require (str): optional, passed on to om.runtime.require()
            reset (bool): optional, if True will resubmit finished jobs
            task_group (str): optional, a specified task group, if not given
               will generate a unique id

        Returns:
            list of started tasks
        """
        nbname = self.jobname
        # generate metadata
        self._generate_jobs(nbname, jobs, job_ids=job_ids, task_group=task_group)
        tasks = self.restart(require=require, reset=reset, task_group=task_group)
        return tasks

    def list(self, task_group=None, **kwargs):
        """
        List all generated tasks for this map call

        Args:
            task_group (str): optional, the task group
            **kwargs (kwargs): kwargs to om.jobs.list()

        Returns:
            list of tasks generated for map call
        """
        om = self.runtime.omega
        task_group = task_group or self.task_group
        nbname = self.jobname
        nbname += f'/{task_group}' if task_group else ''
        return om.jobs.list(f'tasks/{nbname}*', **kwargs)

    def restart(self, task_group=None, reset=False, require=None):
        """
        Run notebook for every entry in tasks/ with no result

        For every entry in om.jobs.list('tasks/*') will check if
        there is a result yet. If not, will call om.runtime.job().run()
        for the given notebook.

        Usage::

            generate_jobs(...)
            restart()

            This will start all jobs that do not have a result yet. To
            start the jobs even if there is a result already, set reset=True.

        Notes:

            ``metadata.attributes['job']`` will record the task status::

            * ``task_id``: the celery task id
            * ``status``: task status, initialized to PENDING
        """
        om = self.runtime.omega
        nbname = self.jobname
        nbname += f'/{task_group}' if task_group else ''
        tasks_nb = om.jobs.list(f'tasks/{nbname}*')
        tasks = []
        for nb in tasks_nb:
            results = om.jobs.list(f'results/{nb}*')
            if not results or reset:
                # generate meta data before running
                meta = om.jobs.metadata(nb)
                job = meta.attributes['job']
                job['status'] = 'pending'
                meta.save()
                # run
                task_rt = om.runtime.require(require).job(nb).run()
                print(f'started {nb} => {task_rt}')
                # update metadata to keep track
                meta = om.jobs.metadata(nb)
                job = meta.attributes['job']
                job['task_id'] = task_rt.id
                meta.save()
                tasks.append(meta)
            else:
                print(f"{nb} has already got results")
        return tasks

    def status(self, task_group=None):
        """
        get the status and the celery id of of each task

        Args:
            task_group (str): the task group id, defaults to None

        Returns::

            pd.DataFrame
                name: task name
                task_id: runtime task id
                status: final status, PENDING, RUNNING, SUCCESS, FAILURE
                run_status: current status, PENDING, RUNNING, SUCCESS, FAILURE
        """
        om = self.runtime.omega
        AsyncResult = self.runtime.celeryapp.AsyncResult
        task_group = task_group or self.task_group
        nbname = self.jobname
        nbname += f'/{task_group}' if task_group else ''
        tasks_nb = om.jobs.list(f'tasks/{nbname}*')
        stats = []
        for nb in tasks_nb:
            meta = om.jobs.metadata(nb)
            task_id = meta.attributes['job'].get('task_id')
            status = AsyncResult(task_id).status if task_id else 'invalid'
            job_runs = meta.attributes.get('job_runs')
            if job_runs:
                run_status = job_runs[-1]['status'] if job_runs else '(waiting)'
            else:
                run_status = 'unknown'
            stats.append((nb, task_id, status, run_status))
        if not tasks_nb:
            print("there are no tasks")
        return pd.DataFrame(stats, columns=['name', 'task_id', 'status', 'run_status'])

    def _generate_jobs(self, nb, jobs, job_ids=None, task_group=None):
        """
        From a notebook, generate a task notebook, parametrized to the jobs arg

        For every job spec in jobs, will generate a copy of the given notebook
        as tasks/{nb}/{task_group}-{id}. The task_group can be set arbitrary or
        be generated as a unique id.

        Once a task is executed, the results will be stored in a new notebook,
        in results/tasks/{nb}/{task_group}-{id}. This contains the full execution
        trace, including graphs and logs of the notebook, and can be accessed
        from Python or Jupyter.

        Usage::

            generate_jobs('main', range(10))

            This will create 10 copies of main, storing each
            in the tasks folder.

            Use restart() to start the tasks on the om runtime

        Args:
            nb (str): the name of the notebook
            jobs (iter): an interable of objects to pass as a job
            job_ids (list): optional, list of job ids. If passed, the
               job id will be used to name the task id, else
               it is the current count
        """
        from nbformat import v4 as nbv4

        om = self.runtime.omega
        job_ids = list(job_ids) if job_ids is not None else None
        self.task_group = task_group = task_group or self._make_task_group()
        tasks = []
        om.jobs.create('#do not delete', 'results/.placeholder')
        om.jobs.create('#do not delete', 'tasks/.placeholder')
        om.jobs.create('#do not delete', 'results/tasks/.placeholder')
        for i, job in enumerate(jobs):
            main_nb = om.jobs.get(nb)
            job_id = i if job_ids is None else job_ids[i]
            task_name = f'tasks/{nb}/{task_group}-{job_id}'
            print(f'generating task {task_name}')
            if om.jobs.metadata(task_name):
                om.jobs.drop(task_name, force=True)
            # store setup of omegaml in main.nb
            job = dict(param=job,
                       job_id=job_id,
                       task_group=task_group,
                       task_name=task_name)
            valid_auth = om.runtime.auth and om.runtime.auth.userid
            auth = (om.runtime.auth.__dict__
                    if valid_auth else dict(userid='', apikey=''))
            code = init_cell_code.format(job=job,
                                         **auth)
            init_cell = nbv4.new_code_cell(source=code)
            done_cell = nbv4.new_code_cell(source=done_cell_code)
            main_nb['cells'].insert(0, init_cell)
            main_nb['cells'].append(done_cell)
            task_meta = om.jobs.put(main_nb, task_name, attributes={'job': job})
            tasks.append(task_meta)

    def _make_task_group(self, max_idlen=8):
        from hashlib import md5
        from uuid import uuid4

        om = self.runtime.omega
        nbname = self.jobname
        value = md5(uuid4().bytes).hexdigest()
        existing = ','.join(om.jobs.list(f'tasks/{nbname}*'))
        while True:
            candidate = value[0:max_idlen]
            if candidate not in existing:
                break
        return candidate


JobTasks.map.__doc__ = JobTasks.__doc__



from unittest import TestCase
from unittest.mock import patch, MagicMock

import numpy as np
from numpy.testing import assert_almost_equal
from pathlib import Path
from sklearn.datasets import make_classification
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import GridSearchCV

from omegaml import Omega
from omegaml.tests.core.cli.scenarios import CliTestScenarios
from omegaml.tests.util import OmegaTestMixin


class CliRuntimeTests(CliTestScenarios, OmegaTestMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.om = Omega()
        self.clean()

    def test_cli_runtime_model_fit_predict(self):
        self.make_model('reg')
        reg = self.om.models.get('reg')
        self.make_dataset_from_dataframe('test', N=10)
        # -- fit the model
        self.cli('runtime model reg fit test[x] test[y]')
        self.assertLogContains('info', 'Metadata')
        # -- run a prediction, we get back actual data
        self.cli('runtime model reg predict test[x]', new_start=True)
        self.assertLogContains('info', '[[')
        self.assertIsInstance(self.get_log('info')[0][0], np.ndarray)
        # -- run a prediction, we store the results
        self.cli('runtime model reg predict test[x] --result predictions')
        #    expect metadata since we ask to write data to predictions
        self.assertLogContains('info', 'Metadata')
        df_expect = self.om.datasets.get('test')['y']
        #    check the data has actually been stored as requested
        df_pred = self.om.datasets.get('predictions')
        self.assertIsNotNone(df_pred)
        self.assertEqual(len(df_expect), len(df_pred))

    def test_cli_runtime_model_gridsearch(self):
        X, y = make_classification()
        logreg = LogisticRegression(solver='liblinear')
        self.om.models.put(logreg, 'logreg')
        # perform gridsearch on runtime
        self.om.datasets.put(X, 'testX')
        self.om.datasets.put(y, 'testY')
        self.cli('runtime model logreg gridsearch testX testY --param C=[.1,.5,1.0]')
        model_meta = self.om.models.metadata('logreg')
        self.assertIn('gridsearch', model_meta.attributes)
        gsmodel_name = model_meta.attributes['gridsearch'][0]['gsModel']
        gsmodel = self.om.models.get(gsmodel_name)
        self.assertIsInstance(gsmodel, GridSearchCV)

    def test_cli_runtime_script_run(self):
        pkgpath = self.get_package_path()
        self.cli(f'scripts put {pkgpath} helloworld')
        self.assertLogContains('info', 'Metadata')
        # run without kwargs
        self.cli('runtime script helloworld run', new_start=True)
        self.assertLogContains('info', 'hello from helloworld')
        # with kwargs
        self.cli('runtime script helloworld run foo=bar', new_start=True)
        self.assertLogContains('info', '"foo": "bar"')

    def test_cli_runtime_model_fit_predict_versions(self):
        self.make_model('reg')
        reg = self.om.models.get('reg')
        self.make_dataset_from_dataframe('test', N=10, m=2, b=0)
        self.make_dataset_from_dataframe('test2', N=10, m=5, b=10)
        # -- fit the model
        self.cli('runtime model reg fit test[x] test[y]')
        self.cli('runtime model reg fit test2[x] test2[y]')
        # -- run a prediction, we store the results, by previous pointer
        self.cli('runtime model reg^ predict test[x] --result pred1')
        self.cli('runtime model reg predict test[x] --result pred2')
        df_expect1 = self.om.datasets.get('test')['y'].values
        df_expect2 = self.om.datasets.get('test2')['y'].values
        #    check the data has actually been stored as requested
        df_pred1 = self.om.datasets.get('pred1').flatten()
        df_pred2 = self.om.datasets.get('pred2').flatten()
        assert_almost_equal(df_expect1, df_pred1, decimal=1)
        assert_almost_equal(df_expect2, df_pred2, decimal=1)

    def test_cli_export_import_compressed(self):
        om = self.om
        self.make_model('reg')
        self.make_dataset_from_dataframe('sample')
        # export into a compressed archive
        self.cli('runtime export --compress --path=/tmp/testcli')
        expfile = self.get_log('info', as_text=True)[-1]
        self.assertTrue(Path(expfile).exists())
        # delete everything, import
        self.clean()
        self.cli(f'runtime import --path={expfile}')
        self.assertIn('reg', om.models.list())
        self.assertIn('sample', om.datasets.list())
        # import and promote model (twice to get 2 versions)
        self.clean()
        self.cli(f'runtime import --promote --path={expfile} models/*')
        self.cli(f'runtime import --promote --path={expfile} models/*')
        self.assertEqual(len(om.models.revisions('reg')), 2)

    def test_cli_export_import_path(self):
        om = self.om
        self.make_model('reg')
        self.make_dataset_from_dataframe('sample')
        # export into a compressed archive
        self.cli('runtime export --path=/tmp/testcli')
        expfile = self.get_log('info', as_text=True)[-1]
        self.assertTrue(Path(expfile).exists())
        # delete everything, import
        self.clean()
        self.cli(f'runtime import --path={expfile}')
        self.assertIn('reg', om.models.list())
        self.assertIn('sample', om.datasets.list())
        # import and promote model (twice to get 2 versions)
        self.clean()
        self.cli(f'runtime import --promote --path={expfile} models/*')
        self.cli(f'runtime import --promote --path={expfile} models/*')
        self.assertEqual(len(om.models.revisions('reg')), 2)

    def test_cli_deploy_steps(self):
        om = self.om
        self.make_model('reg')
        self.make_dataset_from_dataframe('sample')
        # write a deployfile
        deployfile = Path(om.defaults.OMEGA_TMP) / 'deployfile.yaml'
        with open(deployfile, 'w') as fout:
            steps = """
            runtime:
                - action: {action}
                  name: |
                     models/reg 
                     datasets/sample
            """
            fout.writelines(steps)
        # dry run
        self.cli(f'runtime deploy export --steps {deployfile} --dry')
        expected = self.pretend_log('DRY: om runtime  export models/reg datasets/sample')
        self.assertLogContains('info', expected)
        self.cli(f'runtime deploy import --steps {deployfile} --dry')
        expected = self.pretend_log('DRY: om runtime  import models/reg datasets/sample')
        self.assertLogContains('info', expected)
        # actually do it
        self.cli(f'runtime deploy export --steps {deployfile}')
        # delete everything, import
        self.clean()
        self.cli(f'runtime deploy import --steps {deployfile}')
        print(self.get_log('info'))
        self.assertIn('reg', om.models.list())
        self.assertIn('sample', om.datasets.list())
        # override action
        with open(deployfile, 'w') as fout:
            steps = """
                    runtime:
                        - action: foobar
                          name: test
                    """
            fout.writelines(steps)
        # dry run
        self.cli(f'runtime deploy export --steps {deployfile} --dry')
        expected = self.pretend_log('DRY: om runtime foobar test')


    def test_cli_restart_app(self):
        om = self.om
        om.runtime = MagicMock()
        om.runtime.auth.userid = 'testuser'
        om.runtime.auth.apikey = 'apikey'
        self.make_model('reg')
        # use the default REST API as the /apps url
        with patch('omegaml.client.cli.runtime.requests') as requests, \
             patch('omegaml.client.cli.runtime.get_omega') as get_omega:
            get_omega.return_value = om
            self.cli('runtime restart app apps/test')
            requests.get.assert_called()
            self.assertEqual(requests.get.call_args_list[0][0], ('local/apps/api/stop/testuser/test',))
            self.assertEqual(requests.get.call_args_list[1][0], ('local/apps/api/start/testuser/test',))
        # use a specific apphub URL
        with patch('omegaml.client.cli.runtime.requests') as requests, \
             patch('omegaml.client.cli.runtime.get_omega') as get_omega:
            get_omega.return_value = om
            self.cli('runtime restart app apps/test --apphub-url http://myapphub.com')
            requests.get.assert_called()
            self.assertEqual(requests.get.call_args_list[0][0], ('http://myapphub.com/apps/api/stop/testuser/test',))
            self.assertEqual(requests.get.call_args_list[1][0], ('http://myapphub.com/apps/api/start/testuser/test',))




from omegaml.restapi.util import strict, AnyObject
from .app import create_app

# list of regex to filter available resources in format
#   resource/name/action
#   e.g. model/.*/.* => allow all models
#   e.g. model/foo/predict => allow only prediction
# only regex that match at least one regex can succeed
# empty list means all objects and actions are allowed
# see OmegaResourceMixin.check_object_authorization()
resource_filter = []



from unittest import TestCase

import os
import sys
from marshmallow import fields, Schema

from omegaml import Omega, restapi
from omegaml.backends.virtualobj import virtualobj
from omegaml.client.auth import OmegaRestApiAuth
from omegaml.tests.core.restapi.util import RequestsLikeTestClient
from omegaml.tests.util import OmegaTestMixin


class ServiceDirectResourceTests(OmegaTestMixin, TestCase):
    base_url = '/api/service/'

    def setUp(self):
        app = restapi.create_app()
        self.client = RequestsLikeTestClient(app, is_json=True)
        self.om = Omega()
        self.auth = OmegaRestApiAuth('user', 'pass')
        self.clean()

    def tearDown(self):
        pass

    def url(self, pk=None, action=None, query=None):
        url = self.base_url
        # avoid duplicating /service/service
        pk = pk.replace('service/', '')
        if pk is not None:
            url += '{pk}/'.format(**locals())
        if action is not None:
            url += '{action}'.format(**locals())
        if query is not None:
            url += '?{query}'.format(**locals())
        return url

    def test_service_run_script(self):
        """ a service with no specific input and output requirements """
        # essentially just a script, but run with service semantics
        basepath = os.path.join(os.path.dirname(sys.modules['omegaml'].__file__), 'example')
        pkgpath = os.path.abspath(os.path.join(basepath, 'demo', 'helloworld'))
        om = self.om
        pkg = 'pkg://{}'.format(pkgpath)
        # put script
        meta = om.scripts.put(pkg, 'helloworld')
        # run the script on the cluster
        resp = self.client.post(self.url('service/helloworld', query='text=foo'), json={'foo': 'bar'})
        self.assertHttpOK(resp)
        data = self.deserialize(resp)
        # note the json payload is not sent back, because it is not passed as **kwargs but as part of *args
        expected = list(['hello from helloworld', {'text': 'foo', 'pure_python': False}])
        self.assertEqual(data.get('data'), expected)

    def test_service_script_signature_valid(self):
        """ a service with defined data type """
        # essentially just a script, but run with service semantics
        basepath = os.path.join(os.path.dirname(sys.modules['omegaml'].__file__), 'example')
        pkgpath = os.path.abspath(os.path.join(basepath, 'demo', 'helloservice'))
        om = self.om
        pkg = 'pkg://{}'.format(pkgpath)
        om.scripts.put(pkg, 'helloservice')

        # specify service input and output
        class MyInputSchema(Schema):
            factor = fields.Float()

        class MyResultSchema(Schema):
            a = fields.List(fields.Float())
            b = fields.List(fields.Float())

        om.scripts.link_datatype('helloservice', X=MyInputSchema, result=MyResultSchema)
        # can we get documentation that matches the types
        resp = self.client.get(self.url('helloservice', action='doc', query='text=foo'), json={'factor': 5.0})
        self.assertHttpOK(resp)
        specs = resp.json
        self.assertIn('paths', specs)
        self.assertIn('/api/service/helloservice', specs['paths'])
        self.assertIn('definitions', specs)
        self.assertIn('helloservice_X', specs['definitions'])
        self.assertIn('helloservice_result', specs['definitions'])
        # run the script
        resp = self.client.post(self.url('helloservice', query='text=foo'), json={'factor': 5.0})
        self.assertHttpOK(resp)
        data = self.deserialize(resp)
        expected = {'b': [0.0, 5.0, 10.0, 15.0, 20.0], 'a': [0.0, 5.0, 10.0, 15.0, 20.0]}
        self.assertEqual(data, expected)

    def test_service_script_signature_invalid(self):
        """ a service with defined data type """
        # essentially just a script, but run with service semantics
        basepath = os.path.join(os.path.dirname(sys.modules['omegaml'].__file__), 'example')
        pkgpath = os.path.abspath(os.path.join(basepath, 'demo', 'helloservice'))
        om = self.om
        pkg = 'pkg://{}'.format(pkgpath)
        om.scripts.put(pkg, 'helloservice')

        # specify service input and output
        class MyInputSchema(Schema):
            factor = fields.Float()

        class MyResultSchema(Schema):
            a = fields.List(fields.Float())
            b = fields.List(fields.Float())

        om.scripts.link_datatype('helloservice', X=MyInputSchema, result=MyResultSchema)
        # can we get documentation that matches the types
        resp = self.client.get(self.url('service/helloservice', action='doc'))
        self.assertHttpOK(resp)
        specs = resp.json
        self.assertIn('paths', specs)
        self.assertIn('/api/service/helloservice', specs['paths'])
        self.assertIn('definitions', specs)
        self.assertIn('helloservice_X', specs['definitions'])
        self.assertIn('helloservice_result', specs['definitions'])
        # run the script
        resp = self.client.post(self.url('service/helloservice', query='text=foo'), json={'xfactor': 5.0})
        self.assertEqual(resp.status_code, 400)
        expected = {'message': "{'script': 'service/helloservice', 'args': ({'xfactor': 5.0},), "
                               "'kwargs': {'text': 'foo', 'xfactor': 5.0, 'pure_python': False}, "
                               '\'result\': "ValidationError({\'xfactor\': [\'Unknown '
                               'field.\']})", \'runtimes\': 0.028303, \'started\': '
                               "'2022-09-04T12:59:41.180949', 'ended': "
                               "'2022-09-04T12:59:41.209252'}"}
        self.assertTrue('ValidationError' in resp.json['message'])

    def test_service_model_signature_invalid(self):
        """ a service with defined data type """
        # essentially just a script, but run with service semantics
        basepath = os.path.join(os.path.dirname(sys.modules['omegaml'].__file__), 'example')
        pkgpath = os.path.abspath(os.path.join(basepath, 'demo', 'helloservice'))
        om = self.om
        pkg = 'pkg://{}'.format(pkgpath)
        om.models.put(pkg, 'helloservice')

        # specify service input and output
        class MyInputSchema(Schema):
            factor = fields.Float()

        class MyResultSchema(Schema):
            a = fields.List(fields.Float())
            b = fields.List(fields.Float())

        om.models.link_datatype('helloservice', X=MyInputSchema, result=MyResultSchema)
        # can we get documentation that matches the types
        resp = self.client.get(self.url('service/helloservice', action='doc'))
        self.assertHttpOK(resp)
        specs = resp.json
        self.assertIn('paths', specs)
        self.assertIn('/api/service/helloservice', specs['paths'])
        self.assertIn('definitions', specs)
        self.assertIn('helloservice_X', specs['definitions'])
        self.assertIn('helloservice_result', specs['definitions'])
        # run the script
        resp = self.client.post(self.url('service/helloservice', query='text=foo'), json={'xfactor': 5.0})
        self.assertEqual(resp.status_code, 400)
        expected = {'message': "{'script': 'service/helloservice', 'args': ({'xfactor': 5.0},), "
                               "'kwargs': {'text': 'foo', 'xfactor': 5.0, 'pure_python': False}, "
                               '\'result\': "ValidationError({\'xfactor\': [\'Unknown '
                               'field.\']})", \'runtimes\': 0.028303, \'started\': '
                               "'2022-09-04T12:59:41.180949', 'ended': "
                               "'2022-09-04T12:59:41.209252'}"}
        self.assertTrue('ValidationError' in resp.json['message'])

    def test_service_run_virtualobj_script(self):
        om = self.om

        @virtualobj
        def myscript(data=None, method=None, meta=None, store=None, tracking=None, **kwargs):
            return {'data': data, 'method': method}

        om.scripts.put(myscript, 'myscript')
        # check myscript is actually deserialized by runtime
        myscript = None
        # run the script on the cluster
        resp = self.client.post(self.url('service/myscript', action='run', query='text=foo'), json={'foo': 'bar'})
        self.assertHttpOK(resp)
        data = self.deserialize(resp)
        expected = {'data': {'foo': 'bar'}, 'method': 'run'}
        self.assertEqual(data, expected)

    def test_service_predict_virtualobj_model_signature(self):
        om = self.om

        @virtualobj
        def mymodel(data=None, method=None, meta=None, store=None, tracking=None, **kwargs):
            return {'data': data, 'method': method}

        # specify service input and output
        class MyInputSchema(Schema):
            factor = fields.Float()

        class MyResultSchema(Schema):
            a = fields.List(fields.Float())
            b = fields.List(fields.Float())

        om.models.put(mymodel, 'mymodel')
        om.models.link_datatype('mymodel', X=MyInputSchema, result=MyResultSchema, actions=['predict'])
        # check mymodel is actually deserialized by runtime
        mymodel = None
        # run the script on the cluster
        resp = self.client.post(self.url('service/mymodel', action='predict', query='text=foo'), json={'factor': 1.0})
        self.assertHttpOK(resp)
        data = self.deserialize(resp)
        expected = {'data': [{'factor': 1.0}], 'method': 'predict'}
        self.assertEqual(data, expected)
        # run again with invalid signature
        resp = self.client.post(self.url('service/mymodel', action='predict', query='text=foo'), json={'xfactor': 1.0})
        self.assertEqual(resp.status_code, 400)
        self.assertTrue('ValidationError' in resp.json['message'])

    def test_service_predict_virtualobj_model_nosignature(self):
        om = self.om

        @virtualobj
        def mymodel(data=None, method=None, meta=None, store=None, tracking=None, **kwargs):
            return {'data': data, 'method': method}

        om.models.put(mymodel, 'mymodel')
        # check mymodel is actually deserialized by runtime
        mymodel = None
        # run the script on the cluster
        resp = self.client.post(self.url('service/mymodel', action='predict', query='text=foo'), json={'foo': 'bar'})
        self.assertHttpOK(resp)
        data = self.deserialize(resp)
        expected = {'data': [{'foo': 'bar'}], 'method': 'predict'}
        self.assertEqual(data, expected)

    def test_service_run_async(self):
        basepath = os.path.join(os.path.dirname(sys.modules['omegaml'].__file__), 'example')
        pkgpath = os.path.abspath(os.path.join(basepath, 'demo', 'helloworld'))
        om = self.om
        pkg = 'pkg://{}'.format(pkgpath)
        # put script
        meta = om.scripts.put(pkg, 'helloworld')
        # run the script on the cluster
        resp = self.client.post(self.url('service/helloworld', action='run', query='text=foo'),
                                headers=self._async_headers, json={})
        resp = self._check_async(resp)
        self.assertHttpOK(resp)
        data = self.deserialize(resp)['response']
        # since the response is not valid json, the 'data' key is inserted by GenericServiceResource
        expected = list(['hello from helloworld', {'text': 'foo', 'pure_python': False}])
        self.assertEqual(data['data'], expected)


class ServiceV1ResourceTests(ServiceDirectResourceTests):
    base_url = '/api/v1/service/'



import glob

import os
import sys
from setuptools import setup, find_packages

README = open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()
version = open(os.path.join(os.path.dirname(__file__), 'omegaml', 'VERSION')).read()

# extras
tables = ['tables>=3.7']
graph_deps = ['matplotlib~=3.5', 'seaborn~=0.11', 'imageio~=2.6', 'plotext~=1.0']
dashserve_deps = ['dashserve']
sql_deps = ['sqlalchemy', 'ipython-sql']
snowflake_deps = ['snowflake-sqlalchemy==1.2.3']
iotools_deps = ['boto>=2.49.0']
streaming_deps = ['minibatch[all]>=0.5.0']
jupyter_deps = ['jupyterlab', 'jupyterhub==2.2.1']
mlflow_deps = ['mlflow~=1.21']
dev_deps = ['pytest', 'twine', 'flake8', 'mock', 'behave', 'splinter', 'ipdb', 'bumpversion']

# -- tensorflow specifics
#    see https://www.tensorflow.org/install/source
tf_version = os.environ.get('TF_VERSION') or '2.3.1'
tf_match = os.environ.get('TF_VERSION_MATCH', '==')
if tf_version.startswith('1.15'):
    assert sys.version_info[:2] <= (3, 7), "TF < 2.x requires Python <= 3.7"
    tf_deps = ['tensorflow=={}'.format(tf_version)]
    tf_deps = tf_deps + ['tensorflow-gpu==1.15.0', 'h5py==2.10.0']
    keras_deps = ['keras==2.2.4']
elif (3, 8) <= sys.version_info[:2] < (3, 9) :
    major, minor, *_ = (int(v) for v in tf_version.split('.'))
    assert (major, minor) >= (2, 2), "Python version 3.8 only supported by TF >= 2.2"
    tf_deps = ['tensorflow{}{}'.format(tf_match, tf_version)]
    keras_deps = ['keras~=2.4.3']
elif sys.version_info[:2] >= (3, 9):
    major, minor, *_ = (int(v) for v in tf_version.split('.'))
    tf_issue = "https://github.com/tensorflow/tensorflow/issues/44485"
    tf_deps = ['tensorflow>=2.5,<2.9']
    keras_deps = ['keras~=2.4.3']
else:
    # python 3.6, tensorflow 2.3.1 only supported with keras<=2.5
    tf_deps = ['tensorflow{}{}'.format(tf_match, tf_version)]
    keras_deps = ['keras<=2.5']

# all deps
all_deps = (tables + graph_deps + dashserve_deps + sql_deps + iotools_deps
            + streaming_deps + jupyter_deps + snowflake_deps)
client_deps = (tables + dashserve_deps + sql_deps + iotools_deps + streaming_deps)

setup(
    name='omegaml',
    version=version,
    packages=find_packages(),
    include_package_data=True,
    data_files=[
        ('omegaml/docs', glob.glob('./docs/source/nb/*.ipynb')),
        ('omegaml/runtimes/rsystem', glob.glob('./runtimes/rsystem/*.R')),
    ],
    scripts=glob.glob('./scripts/runtime/*'),
    license='Apache 2.0 + "No Sell, Consulting Yes" License Condition',
    description='An open source DataOps, MLOps platform for humans',
    long_description=README,
    long_description_content_type='text/x-rst',
    url='https://omegaml.io/',
    author='Patrick Senti',
    author_email='patrick.senti@omegaml.io',
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Development Status :: 4 - Beta',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Operating System :: POSIX :: Linux',
        'License :: OSI Approved :: Apache Software License',
    ],
    install_requires=[
        'celery>4.4,<5.0',
        'importlib-metadata<5.0', # due to https://github.com/celery/kombu/pull/1601
        'joblib>=0.9.4',
        'jupyter-client>=4.1.1',
        'mongoengine~=0.24.1',
        'pymongo~=4.0.2',  # mongoengine 0.24.1 compatibility
        'pandas>1.1',
        'numpy>=1.16.4',
        'scipy>=0.17.0',
        'scikit-learn>=0.21',
        'PyYAML>=3.12',
        'flask-restx>=0.4.0',
        'werkzeug<2.1.0', # due to flask-restx, https://github.com/python-restx/flask-restx/issues/422
        'markupsafe==2.0.1', # due to flask/markupsafe, https://github.com/pallets/markupsafe/issues/284
        'croniter>=0.3.30',
        'nbformat>=4.0.1',
        'nbconvert>=6.4.0',
        'dill>=0.3.2',
        'callable-pip>=1.0.0',
        'appdirs>=1.4.3',
        'cron-descriptor>=1.2.31',
        'docopt>=0.6.2',
        'requests>=2.20.0',
        # fix tensorflow pulling wrong version of absl-py,
        # https://github.com/tensorflow/tensorflow/issues/26691#issuecomment-525519742
        'absl-py>=0.8.1',
        'tqdm>=4.32.2',
        'honcho>=1.0.1',  # not strictly required, but used in docker compose
        'tabulate>=0.8.2',  # required in cli
        'smart_open', # required in cli
        'imageio>=2.3.0', # require to store images
        'psutil>=5.8', # required for profiling tracker
        'cachetools>=5.0.0', # required for session caching
        'apispec>=5.2.2', # required for openapi generation
        'marshmallow>=3.17.0', # required for openapi generation
    ],
    extras_require={
        'graph': graph_deps,
        'tables': tables,
        'tensorflow': tf_deps,
        'keras': keras_deps,
        'jupyter': jupyter_deps,
        'dashserve': dashserve_deps,
        'sql': sql_deps,
        'snowflake': snowflake_deps,
        'mlflow': mlflow_deps,
        'iotools': iotools_deps,
        'streaming': streaming_deps,
        'all': all_deps,
        'client': client_deps,
        'all-client': client_deps,
        'dev': dev_deps,
    },
    entry_points={
        'console_scripts': ['om=omegaml.client.cli:climain'],
    }
)




from omegaml.client.util import get_omega


class StoresCommandMixin:
    """ Common commands for OmegaStore related commands

    Usage:
        class DatasetsCommand(StoresCommandMixin, CommandBase:
            command = 'datasets'

        This will automatically provide list, drop, metadata
        commands for the store given by the command variable
    """
    command = 'unspecified'

    def put(self):
        raise NotImplementedError()

    def get(self):
        raise NotImplementedError()

    def list(self):
        om = get_omega(self.args)
        pattern = self.args.get('<pattern>')
        regexp = self.args.get('--regexp') or self.args.get('-E')
        raw = self.args.get('--raw')
        hidden = self.args.get('--hidden')
        kwargs = dict(regexp=pattern) if regexp else dict(pattern=pattern)
        store = getattr(om, self.command)
        self.logger.info(store.list(raw=raw, hidden=hidden, **kwargs))

    def drop(self):
        om = get_omega(self.args)
        name = self.args.get('<name>')
        store = getattr(om, self.command)
        self.logger.info(store.drop(name))

    def metadata(self):
        om = get_omega(self.args)
        name = self.args.get('<name>')
        store = getattr(om, self.command)
        self.logger.info(store.metadata(name).to_json())

    def plugins(self):
        om = get_omega(self.args)
        store = getattr(om, self.command)
        for kind, plugincls in store.defaults.OMEGA_STORE_BACKENDS.items():
            self.logger.info(kind, plugincls.__doc__)

    def mixins(self):
        om = get_omega(self.args)
        store = getattr(om, self.command)
        for kind, plugincls in store.defaults.OMEGA_STORE_MIXINS.items():
            self.logger.info(kind, plugincls.__doc__)



from __future__ import absolute_import

import unittest
from unittest import TestCase

import numpy as np
import os
import pandas as pd
import sys
from numpy.testing import assert_array_almost_equal
from sklearn.datasets import make_classification
from sklearn.exceptions import NotFittedError
from sklearn.linear_model import LinearRegression
from sklearn.linear_model import LogisticRegression
from sklearn.linear_model import SGDRegressor
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import GridSearchCV
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import DataConversionWarning

from omegaml import Omega
from omegaml.backends.virtualobj import virtualobj
from omegaml.tests.util import OmegaTestMixin
from omegaml.util import delete_database, reshaped


class RuntimeTests(OmegaTestMixin, TestCase):

    def setUp(self):
        TestCase.setUp(self)
        om = self.om = Omega()
        self.clean()
        self.clean(bucket='test')

    def tearDown(self):
        TestCase.tearDown(self)

    def test_predict(self):
        # create some data
        x = np.array(list(range(0, 10)))
        y = x * 2
        df = pd.DataFrame({'x': x,
                           'y': y})
        X = df[['x']]
        Y = df[['y']]
        # put into Omega
        os.environ['DJANGO_SETTINGS_MODULE'] = ''
        om = Omega()
        om.runtime.celeryapp.conf.CELERY_ALWAYS_EAGER = True
        om.datasets.put(X, 'datax')
        om.datasets.put(Y, 'datay')
        om.datasets.get('datax')
        om.datasets.get('datay')
        # create a model locally, fit it, store in Omega
        lr = LinearRegression()
        lr.fit(X, Y)
        pred = lr.predict(X)
        om.models.put(lr, 'mymodel')
        self.assertIn('mymodel', om.models.list('*'))
        # have Omega predict it
        # -- using data already in Omega
        result = om.runtime.model('mymodel').predict('datax')
        pred1 = result.get()
        # -- using data provided locally
        #    note this is the same as
        #        om.datasets.put(X, 'foo')
        #        om.runtimes.model('mymodel').predict('foo')
        result = om.runtime.model('mymodel').predict(X)
        pred2 = result.get()
        self.assertTrue(
            (pred == pred1).all(), "runtimes prediction is different(1)")
        self.assertTrue(
            (pred == pred2).all(), "runtimes prediction is different(2)")

    def test_fit(self):
        # create some data
        x = np.array(list(range(0, 10)))
        y = x * 2
        df = pd.DataFrame({'x': x,
                           'y': y})
        X = df[['x']]
        Y = df[['y']]
        # put into Omega
        os.environ['DJANGO_SETTINGS_MODULE'] = ''
        om = Omega()
        om.runtime.celeryapp.conf.CELERY_ALWAYS_EAGER = True
        om.datasets.put(X, 'datax')
        om.datasets.put(Y, 'datay')
        om.datasets.get('datax')
        om.datasets.get('datay')
        # create a model locally, store (unfitted) in Omega
        lr = LinearRegression()
        om.models.put(lr, 'mymodel2')
        self.assertIn('mymodel2', om.models.list('*'))
        # predict locally for comparison
        lr.fit(X, Y)
        pred = lr.predict(X)
        # try predicting without fitting
        with self.assertRaises(NotFittedError):
            result = om.runtime.model('mymodel2').predict('datax')
            result.get()
        # have Omega fit the model then predict
        result = om.runtime.model('mymodel2').fit('datax', 'datay')
        result.get()
        # check the new model version metadata includes the datax/y references
        meta = om.models.metadata('mymodel2')
        self.assertIn('Xmeta', meta.attributes['dataset'])
        self.assertIn('Ymeta', meta.attributes['dataset'])
        # -- using data already in Omega
        result = om.runtime.model('mymodel2').predict('datax')
        pred1 = result.get()
        # -- using data provided locally
        #    note this is the same as
        #        om.datasets.put(X, 'foo')
        #        om.runtimes.model('mymodel2').predict('foo')
        result = om.runtime.model('mymodel2').fit(X, Y)
        result = om.runtime.model('mymodel2').predict(X)
        pred2 = result.get()
        # -- check the local data provided to fit was stored as intended
        meta = om.models.metadata('mymodel2')
        self.assertIn('Xmeta', meta.attributes['dataset'])
        self.assertIn('Ymeta', meta.attributes['dataset'])
        self.assertIn('_fitX', meta.attributes['dataset'].get('Xmeta').get('name'))
        self.assertIn('_fitY', meta.attributes['dataset'].get('Ymeta').get('name'))
        self.assertTrue(
            (pred == pred1).all(), "runtimes prediction is different(1)")
        self.assertTrue(
            (pred == pred2).all(), "runtimes prediction is different(2)")

    def test_partial_fit(self):
        # create some data
        x = np.array(list(range(0, 10)))
        y = x * 2
        df = pd.DataFrame({'x': x,
                           'y': y})
        X = df[['x']][0:2]
        Y = df[['y']][0:2]
        # put into Omega
        os.environ['DJANGO_SETTINGS_MODULE'] = ''
        om = Omega()
        om.runtime.celeryapp.conf.CELERY_ALWAYS_EAGER = True
        om.datasets.put(df[['x']], 'datax-full')
        om.datasets.put(X, 'datax')
        om.datasets.put(Y, 'datay')
        om.datasets.get('datax')
        om.datasets.get('datay')
        # create a model locally, store (unfitted) in Omega
        # -- ignore warnings on y shape
        import warnings
        warnings.filterwarnings("ignore", category=DataConversionWarning)
        lr = SGDRegressor(max_iter=1000, tol=1e-3)
        om.models.put(lr, 'mymodel2')
        # have Omega fit the model to get a start, then predict
        result = om.runtime.model('mymodel2').fit('datax', 'datay')
        result.get()
        # check the new model version metadata includes the datax/y references
        result = om.runtime.model('mymodel2').predict('datax-full')
        pred1 = result.get()
        mse = mean_squared_error(df.y, pred1)
        self.assertGreater(mse, 40)
        # fit mini batches add better training data, update model
        batch_size = 2
        for i, start in enumerate(range(0, len(df))):
            previous_mse = mse
            X = df[['x']][start:start + batch_size]
            Y = df[['y']][start:start + batch_size]
            om.datasets.put(X, 'datax-update', append=False)
            om.datasets.put(Y, 'datay-update', append=False)
            result = om.runtime.model('mymodel2').partial_fit(
                'datax-update', 'datay-update')
            result.get()
            # check the new model version metadata includes the datax/y
            # references
            result = om.runtime.model('mymodel2').predict('datax-full')
            pred1 = result.get()
            mse = mean_squared_error(df.y, pred1)
            self.assertLess(mse, previous_mse)

    def test_partial_fit_chunked(self):
        # create some data
        x = np.array(list(range(0, 100)))
        y = x * 2
        df = pd.DataFrame({'x': x,
                           'y': y})
        # put into Omega
        os.environ['DJANGO_SETTINGS_MODULE'] = ''
        om = Omega()
        om.runtime.celeryapp.conf.CELERY_ALWAYS_EAGER = True
        # generate a large dataset
        for i in range(100):
            om.datasets.put(df, 'data', append=(i > 0))
        # create a model locally, store (unfitted) in Omega
        # -- ignore warnings on y shape
        import warnings
        warnings.filterwarnings("ignore", category=DataConversionWarning)
        lr = SGDRegressor(max_iter=1000, tol=1e-3, random_state=42)
        om.models.put(lr, 'mymodel2')
        # have Omega fit the model to get a start, then predict
        result = om.runtime.model('mymodel2').fit(df[['x']], df[['y']])
        result.get()
        # check the new model version metadata includes the datax/y references
        result = om.runtime.model('mymodel2').predict('data[x]')
        pred1 = result.get()
        mse = mean_squared_error(om.datasets.get('data[y]'), pred1)
        self.assertGreater(mse, 40)
        # fit mini batches add better training data, update model
        result = om.runtime.model('mymodel2').partial_fit('data[x]#', 'data[y]#')
        result = om.runtime.model('mymodel2').predict('data[x]')
        pred1 = result.get()
        mse_2 = mean_squared_error(om.datasets.get('data[y]'), pred1)
        self.assertLess(mse_2, mse)

    def test_predict_pure_python(self):
        # create some data
        x = np.array(list(range(0, 10)))
        y = x * 2
        df = pd.DataFrame({'x': x,
                           'y': y}).astype('O')
        X = [[x] for x in list(df.x)]
        Y = [[y] for y in list(df.y)]
        # put into Omega -- assume a client with pandas, scikit learn
        os.environ['DJANGO_SETTINGS_MODULE'] = ''
        om = Omega()
        om.runtime.pure_python = True
        om.runtime.celeryapp.conf.CELERY_ALWAYS_EAGER = True
        om.datasets.put(X, 'datax')
        om.datasets.put(Y, 'datay')
        Xhat = om.datasets.get('datax')
        Yhat = om.datasets.get('datay')
        self.assertEqual(X, Xhat)
        self.assertEqual(Y, Yhat)
        # have Omega fit the model then predict
        lr = LinearRegression()
        lr.fit(X, Y)
        pred = lr.predict(X)
        om.models.put(lr, 'mymodel2')
        # -- using data provided locally
        #    note this is the same as
        #        om.datasets.put(X, 'foo')
        #        om.runtimes.model('mymodel2').predict('foo')
        result = om.runtime.model('mymodel2').predict(reshaped(X))
        pred2 = result.get()
        self.assertTrue(
            (pred == pred2).all(), "runtimes prediction is different(1)")
        self.assertTrue(
            (pred == pred2).all(), "runtimes prediction is different(2)")

    def test_predict_hdf_dataframe(self):
        # create some data
        x = np.array(list(range(0, 10)))
        y = x * 2
        df = pd.DataFrame({'x': x,
                           'y': y})
        X = df['x']
        Y = df['y']
        # put into Omega -- assume a client with pandas, scikit learn
        os.environ['DJANGO_SETTINGS_MODULE'] = ''
        om = Omega()
        om.runtime.pure_python = True
        om.runtime.celeryapp.conf.CELERY_ALWAYS_EAGER = True
        om.datasets.put(X, 'datax', as_hdf=True)
        om.datasets.put(Y, 'datay', as_hdf=True)
        # have Omega fit the model then predict
        lr = LinearRegression()
        lr.fit(reshaped(X), reshaped(Y))
        pred = lr.predict(reshaped(X))
        om.models.put(lr, 'mymodel2')
        # -- using data provided locally
        #    note this is the same as
        #        om.datasets.put(X, 'foo')
        #        om.runtimes.model('mymodel2').predict('foo')
        result = om.runtime.model('mymodel2').predict('datax')
        pred2 = result.get()
        self.assertTrue(
            (pred == pred2).all(), "runtimes prediction is different(1)")
        self.assertTrue(
            (pred == pred2).all(), "runtimes prediction is different(2)")

    def test_fit_pipeline(self):
        # create some data
        x = np.array(list(range(0, 10)))
        y = x * 2
        df = pd.DataFrame({'x': x,
                           'y': y})
        X = df[['x']]
        Y = df[['y']]
        # put into Omega
        os.environ['DJANGO_SETTINGS_MODULE'] = ''
        om = Omega()
        om.runtime.celeryapp.conf.CELERY_ALWAYS_EAGER = True
        om.datasets.put(X, 'datax')
        om.datasets.put(Y, 'datay')
        om.datasets.get('datax')
        om.datasets.get('datay')
        # create a pipeline locally, store (unfitted) in Omega
        p = Pipeline([
            ('lr', LinearRegression()),
        ])
        om.models.put(p, 'mymodel2')
        self.assertIn('mymodel2', om.models.list('*'))
        # predict locally for comparison
        p.fit(reshaped(X), reshaped(Y))
        pred = p.predict(reshaped(X))
        # have Omega fit the model then predict
        result = om.runtime.model('mymodel2').fit('datax', 'datay')
        result.get()
        result = om.runtime.model('mymodel2').predict('datax')
        pred1 = result.get()
        self.assertTrue(
            (pred == pred1).all(), "runtimes prediction is different(1)")

    def test_score(self):
        # create some data
        x = np.array(list(range(0, 10)))
        y = x * 2
        df = pd.DataFrame({'x': x,
                           'y': y})
        X = df[['x']]
        Y = df[['y']]
        # put into Omega
        os.environ['DJANGO_SETTINGS_MODULE'] = ''
        om = Omega()
        om.runtime.celeryapp.conf.CELERY_ALWAYS_EAGER = True
        om.datasets.put(X, 'datax')
        om.datasets.put(Y, 'datay')
        om.datasets.get('datax')
        om.datasets.get('datay')
        # create a model locally, fit it, store in Omega
        lr = LinearRegression()
        lr.fit(X, Y)
        scores = lr.score(X, Y)
        om.models.put(lr, 'mymodel')
        # fit in omegaml
        r_scores = om.runtime.model('mymodel').score('datax', 'datay').get()
        self.assertEqual(scores, r_scores)

    def test_gridsearch(self):
        X, y = make_classification()
        logreg = LogisticRegression(solver='liblinear')
        os.environ['DJANGO_SETTINGS_MODULE'] = ''
        om = Omega()
        om.runtime.celeryapp.conf.CELERY_ALWAYS_EAGER = True
        om.models.put(logreg, 'logreg')
        params = {
            'C': [0.1, 0.5, 1.0]
        }
        # gridsearch on runtimes
        om.runtime.model('logreg').gridsearch(X, y, parameters=params).get()
        meta = om.models.metadata('logreg')
        # check gridsearch was saved
        self.assertIn('gridsearch', meta.attributes)
        self.assertEqual(len(meta.attributes['gridsearch']), 1)
        self.assertIn('gsModel', meta.attributes['gridsearch'][0])
        # check we can get back the gridsearch model
        gs_model = om.models.get(meta.attributes['gridsearch'][0]['gsModel'])
        self.assertIsInstance(gs_model, GridSearchCV)

    def test_gridsearch_iris(self):
        om = Omega()
        from sklearn.datasets import load_iris
        X, y = load_iris(return_X_y=True)
        df = pd.DataFrame(X)
        df['y'] = y
        om.datasets.put(df, 'iris', append=False)
        from sklearn.cluster import KMeans
        model = KMeans(n_clusters=8)
        # fit & predict remote
        om.models.drop('iris-model', True)
        om.models.put(model, 'iris-model')
        om.runtime.model('iris-model').fit(X, y).get()
        params = {
            'n_clusters': range(1, 8),
        }
        om.runtime.model('iris-model').gridsearch('iris[^y]', 'iris[y]', parameters=params).get()

    def test_ping(self):
        om = Omega()
        result = om.runtime.ping(fox='bar')
        self.assertIn('message', result)
        self.assertIn('worker', result)
        self.assertEqual(result['kwargs'], dict(fox='bar'))

    def test_task_sequence(self):
        om = Omega()
        df = pd.DataFrame({'x': range(1, 10),
                           'y': range(5, 14)})
        lr = LinearRegression()
        om.datasets.put(df, 'sample')
        om.models.put(lr, 'regmodel')
        with om.runtime.sequence() as ctr:
            ctr.ping(wait=False)
            ctr.model('regmodel').fit('sample[x]', 'sample[y]')
            ctr.model('regmodel').predict('sample[x]')
            result = ctr.run()

        data = result.get()
        assert_array_almost_equal(df['y'].values, data[:, 0])

    def test_task_parallel(self):
        om = Omega()
        df = pd.DataFrame({'x': range(1, 10),
                           'y': range(5, 14)})
        lr = LinearRegression()
        om.datasets.put(df, 'sample')
        om.models.put(lr, 'regmodel')
        om.runtime.model('regmodel').fit('sample[x]', 'sample[y]').get()
        with om.runtime.parallel() as ctr:
            ctr.model('regmodel').predict('sample[x]')
            ctr.model('regmodel').predict('sample[x]')
            result = ctr.run()

        data = result.get()
        assert_array_almost_equal(df['y'].values, data[0][:, 0])
        assert_array_almost_equal(df['y'].values, data[1][:, 0])

    def test_task_mapreduce_virtualfn(self):
        om = Omega()
        df = pd.DataFrame({'x': range(1, 10),
                           'y': range(5, 14)})
        lr = LinearRegression()
        om.datasets.put(df, 'sample')
        om.models.put(lr, 'regmodel')
        om.runtime.model('regmodel').fit('sample[x]', 'sample[y]').get()

        @virtualobj
        def combined(data=None, method=None, meta=None, store=None, **kwargs):
            # data is the list of results from the previous tasks
            # we scale results to verify combined actually runs
            return [y * 5 for y in data]

        om.models.put(combined, 'combined')
        with om.runtime.mapreduce() as ctr:
            # two tasks to map
            ctr.model('regmodel').predict('sample[x]')
            ctr.model('regmodel').predict('sample[x]')
            # one task to reduce
            ctr.model('combined').reduce()
            result = ctr.run()

        data = result.get()
        assert_array_almost_equal(df['y'].values * 5, data[0][:, 0])
        assert_array_almost_equal(df['y'].values * 5, data[1][:, 0])

        @virtualobj
        def combined(data=None, method=None, meta=None, store=None, **kwargs):
            # data is the list of results from the previous tasks
            # we return only one result, simulating selection
            return data[0][:, 0]

        om.models.put(combined, 'combined', replace=True)
        with om.runtime.mapreduce() as ctr:
            # two tasks to map
            ctr.model('regmodel').predict('sample[x]')
            ctr.model('regmodel').predict('sample[x]')
            # one task to reduce
            ctr.model('combined').reduce()
            result = ctr.run()

        data = result.get()
        assert_array_almost_equal(df['y'].values, data)

    def test_task_mapreduce_script(self):
        om = Omega()
        df = pd.DataFrame({'x': range(1, 10),
                           'y': range(5, 14)})
        lr = LinearRegression()
        om.datasets.put(df, 'sample')
        om.models.put(lr, 'regmodel')
        om.runtime.model('regmodel').fit('sample[x]', 'sample[y]').get()

        om = Omega()
        basepath = os.path.join(os.path.dirname(sys.modules['omegaml'].__file__), 'example')
        pkgpath = os.path.abspath(os.path.join(basepath, 'demo', 'callback'))
        pkgsrc = 'pkg://{}'.format(pkgpath)
        om.scripts.put(pkgsrc, 'callback')
        with om.runtime.mapreduce() as ctr:
            # two tasks to map
            ctr.model('regmodel').predict('sample[x]')
            ctr.model('regmodel').predict('sample[x]')
            # one task to reduce
            ctr.script('callback').run(as_callback=True)
            result = ctr.run()

        result.get()
        self.assertEqual(len(om.datasets.get('callback_results')), 18)

        with om.runtime.mapreduce() as ctr:
            # two tasks to map
            ctr.model('regmodel').predict('sample[x]')
            # one task to reduce
            ctr.script('callback').run(as_callback=True)
            result = ctr.run()

        result.get()
        self.assertEqual(len(om.datasets.get('callback_results')), 27)

    def test_task_callback(self):
        om = Omega()
        basepath = os.path.join(os.path.dirname(sys.modules['omegaml'].__file__), 'example')
        pkgpath = os.path.abspath(os.path.join(basepath, 'demo', 'callback'))
        pkgsrc = 'pkg://{}'.format(pkgpath)
        om.scripts.put(pkgsrc, 'callback')
        df = pd.DataFrame({'x': range(1, 10),
                           'y': range(5, 14)})
        lr = LinearRegression()
        lr.fit(df[['x']], df['y'])
        om.datasets.put(df, 'sample')
        om.models.put(lr, 'regmodel')
        result = (om.runtime
                  .callback('callback')
                  .model('regmodel')
                  .predict('sample[x]')
                  .get())
        self.assertEqual(len(om.datasets.get('callback_results')), 1)
        result = (om.runtime
                  .callback('callback')
                  .model('regmodel')
                  .predict('sample[x]')
                  .get())
        self.assertEqual(len(om.datasets.get('callback_results')), 2)

    def test_task_callback_bucket(self):
        om = Omega()
        omb = om['test']
        basepath = os.path.join(os.path.dirname(sys.modules['omegaml'].__file__), 'example')
        pkgpath = os.path.abspath(os.path.join(basepath, 'demo', 'callback'))
        pkgsrc = 'pkg://{}'.format(pkgpath)
        omb.scripts.put(pkgsrc, 'callback')
        df = pd.DataFrame({'x': range(1, 10),
                           'y': range(5, 14)})
        lr = LinearRegression()
        lr.fit(df[['x']], df['y'])
        omb.datasets.put(df, 'sample')
        omb.models.put(lr, 'regmodel')
        result = (omb.runtime
                  .callback('callback')
                  .model('regmodel')
                  .predict('sample[x]')
                  .get())
        self.assertEqual(len(omb.datasets.get('callback_results')), 1)
        result = (omb.runtime
                  .callback('callback')
                  .model('regmodel')
                  .predict('sample[x]')
                  .get())
        self.assertEqual(len(omb.datasets.get('callback_results')), 2)

    def test_task_logging(self):
        """ test task python output can be logged per-request """
        om = Omega()
        om.logger.reset()
        # no python logging, only om.logger
        om.runtime.ping(fox='bar', logging=False)
        self.assertEqual(len(om.logger.dataset.get(level='INFO')), 0)
        # python log capture, we get om.logger, omegaml + stdout log
        om.logger.reset()
        om.runtime.mode(logging=True).ping(fox='bar')
        self.assertEqual(len(om.logger.dataset.get(level='INFO')), 4)
        # specific python logger, we get om.logger + celery
        om.logger.reset()
        om.runtime.mode(logging='celery').ping(fox='bar')
        self.assertEqual(len(om.logger.dataset.get(level='INFO')), 3)
        # request a different level, we get celery + stdout
        om.logger.reset()
        om.runtime.mode(logging=('celery', 'DEBUG')).ping(fox='bar')
        self.assertEqual(len(om.logger.dataset.get()), 4)

    def test_task_logging_bucket(self):
        """ test task python output can be logged per-request """
        om = Omega()['test']
        om.logger.reset()
        # no python logging, only om.logger
        om.runtime.ping(fox='bar', logging=False)
        self.assertEqual(len(om.logger.dataset.get(level='INFO')), 0)
        # python log capture, we get om.logger, omegaml + stdout log
        om.logger.reset()
        om.runtime.mode(logging=True).ping(fox='bar')
        self.assertEqual(len(om.logger.dataset.get(level='INFO')), 4)
        # specific python logger, we get om.logger, celery + stdout log
        om.logger.reset()
        om.runtime.mode(logging='celery').ping(fox='bar')
        self.assertEqual(len(om.logger.dataset.get(level='INFO')), 3)
        # request a different level, we get celery + stdout
        om.logger.reset()
        om.runtime.mode(logging=('celery', 'DEBUG')).ping(fox='bar')
        self.assertEqual(len(om.logger.dataset.get()), 4)

    def test_logging_mode(self):
        """ test task python output can be logged for all requests """
        om = Omega()
        om.logger.reset()
        # -- request logging
        om.runtime.mode(local=True, logging=True)
        om.runtime.ping(fox='bar')
        self.assertEqual(len(om.logger.dataset.get(level='INFO')), 4)
        # -- switch off logging
        om.logger.reset()
        om.runtime.mode(local=True, logging=False)
        om.runtime.ping(fox='bar')
        self.assertEqual(len(om.logger.dataset.get(level='INFO')), 0)
        # -- request specific logger
        om.logger.reset()
        om.runtime.mode(local=True, logging=('celery', 'DEBUG'))
        om.runtime.ping(fox='bar')
        self.assertEqual(len(om.logger.dataset.get(level='DEBUG')), 3)

    def test_list_labels(self):
        om = Omega()
        labels = om.runtime.mode(local=True).labels()
        self.assertIsInstance(labels, dict)
        self.assertEqual(['local'], list(labels.values())[0])

    def test_parallel_getall(self):
        om = Omega()
        code = """print('hello')"""
        om.jobs.create(code, 'myjob')
        # --parallel
        with om.runtime.parallel() as crt:
            for i in range(5):
                om.runtime.job(f'myjob').run()
            result = crt.run()
        results = result.getall()
        self.assertEqual(len(results), 5)
        # --sequence
        with om.runtime.sequence() as crt:
            for i in range(5):
                om.runtime.job(f'myjob').run()
            result = crt.run()
        results = result.getall()
        self.assertEqual(len(results), 5)

    @unittest.skip("fails due to job not supporting callbacks")
    def test_mapreduce_getall(self):
        om = Omega()
        code = """print('hello')"""
        om.jobs.create(code, 'myjob')
        # --mapreduce
        with om.runtime.mapreduce() as crt:
            for i in range(5):
                om.runtime.job(f'myjob').run()
            result = crt.run()
        results = result.getall()
        self.assertEqual(len(results), 5)

    def test_predict_multiple_samples(self):
        om = Omega()
        reg = LinearRegression()
        df = pd.DataFrame({'x': range(10)})
        df['y'] = df['x'] * 2 + 3
        reg.fit(df[['x']], df['y'])
        om.models.put(reg, 'regmodel')
        result = om.runtime.model('regmodel').predict([[5], [6]]).get()






from unittest import TestCase

import os
import sys

from omegaml import Omega, restapi
from omegaml.backends.virtualobj import virtualobj
from omegaml.client.auth import OmegaRestApiAuth
from omegaml.tests.core.restapi.util import RequestsLikeTestClient
from omegaml.tests.util import OmegaTestMixin


class ScriptResourceTests(OmegaTestMixin, TestCase):
    def setUp(self):
        app = restapi.create_app()
        self.client = RequestsLikeTestClient(app, is_json=True)
        self.om = Omega()
        self.auth = OmegaRestApiAuth('user', 'pass')
        self.clean()

    def tearDown(self):
        pass

    def url(self, pk=None, action=None, query=None):
        url = '/api/v1/script/'
        if pk is not None:
            url += '{pk}'.format(**locals())
        if action is not None:
            url += '/{action}'.format(**locals())
        if query is not None:
            url += '?{query}'.format(**locals())
        return url

    def test_script_run(self):
        basepath = os.path.join(os.path.dirname(sys.modules['omegaml'].__file__), 'example')
        pkgpath = os.path.abspath(os.path.join(basepath, 'demo', 'helloworld'))
        om = self.om
        pkg = 'pkg://{}'.format(pkgpath)
        # put script
        meta = om.scripts.put(pkg, 'helloworld')
        # run the script on the cluster
        resp = self.client.post(self.url('helloworld', action='run', query='text=foo'), json={'foo': 'bar'})
        self.assertHttpOK(resp)
        data = self.deserialize(resp)
        self.assertIn('runtimes', data)
        self.assertIn('result', data)
        # note the json payload is not part of the response
        expected = list(['hello from helloworld', {'text': 'foo', 'pure_python': False}])
        self.assertEqual(data['result'], expected)

    def test_script_run_virtualobj(self):
        om = self.om

        @virtualobj
        def myscript(data=None, method=None, meta=None, store=None, tracking=None, **kwargs):
            return {'data': data, 'method': method}

        om.scripts.put(myscript, 'myscript')
        # check myscript is actually deserialized by runtime
        myscript = None
        # run the script on the cluster
        resp = self.client.post(self.url('myscript', action='run', query='text=foo'), json={'foo': 'bar'})
        self.assertHttpOK(resp)
        data = self.deserialize(resp)
        self.assertIn('runtimes', data)
        self.assertIn('result', data)
        expected = {'data': {'foo': 'bar'}, 'method': 'run'}
        self.assertEqual(data['result'], expected)

    def test_script_run_async(self):
        basepath = os.path.join(os.path.dirname(sys.modules['omegaml'].__file__), 'example')
        pkgpath = os.path.abspath(os.path.join(basepath, 'demo', 'helloworld'))
        om = self.om
        pkg = 'pkg://{}'.format(pkgpath)
        # put script
        meta = om.scripts.put(pkg, 'helloworld')
        # run the script on the cluster
        resp = self.client.post(self.url('helloworld', action='run', query='text=foo'),
                                headers=self._async_headers, json={})
        resp = self._check_async(resp)
        self.assertHttpOK(resp)
        data = self.deserialize(resp)['response']
        self.assertIn('runtimes', data)
        self.assertIn('result', data)
        expected = list(['hello from helloworld', {'text': 'foo', 'pure_python': False}])
        self.assertEqual(data['result'], expected)
